"""Tests for :py:mod:`katsdpimager.loader_katdal`."""

from collections import defaultdict
import logging
import tempfile
from unittest import mock

import astropy.units as u
from astropy.coordinates import Angle
import astropy.table

import numpy as np
import katpoint
import katsdptelstate

import katsdpmodels.rfi_mask
import katsdpmodels.band_mask

import katdal
from katdal import VisibilityDataV4
from katdal.datasources import TelstateDataSource
from katdal.chunkstore_npy import NpyFileChunkStore
from katdal.test.test_datasources import make_fake_data_source

from nose.tools import (
    assert_equal, assert_true, assert_false, assert_is_instance, assert_in,
    assert_logs, assert_raises
)

from ..loader_katdal import LoaderKatdal
from .. import polarization


ANTENNAS = [
    katpoint.Antenna('m000, -30:42:39.8, 21:26:38.0, 1086.6, 13.5, -8.264 -207.29 8.5965 212.6695 212.6695 1.0, 0:03:13.3 0 0:00:29.9 0:00:45.1 0:00:08.9 -0:00:12.5 0:13:33.4 0:01:18.0, 1.22'),      # noqa: E501
    katpoint.Antenna('m001, -30:42:39.8, 21:26:38.0, 1086.6, 13.5, 1.1205 -171.762 8.4705 209.996 209.996 1.0, -0:40:31.1 0 0:00:05.8 -0:00:18.3 -0:00:02.8 -0:00:42.2 -0:35:36.2 0:01:13.0, 1.22'),   # noqa: E501
    katpoint.Antenna('m002, -30:42:39.8, 21:26:38.0, 1086.6, 13.5, -32.113 -224.236 8.6445 211.7145 211.7145 1.0, 0:43:31.2 0 0:00:42.4 0:05:00.9 0:00:07.4 -0:00:02.1 0:02:16.3 0:01:01.6, 1.22'),    # noqa: E501
    katpoint.Antenna('m003, -30:42:39.8, 21:26:38.0, 1086.6, 13.5, -66.518 -202.276 8.285 213.4215 213.4215 1.0, 1:35:27.0 0 -0:00:53.3 -0:06:05.9 -0:00:00.5 -0:00:07.1 0:33:51.5 0:00:51.3, 1.22')   # noqa: E501
]
TARGET = katpoint.Target('PKS 1934-63, radec, 19:39:25.03, -63:42:45.7, (200.0 12000.0 -11.11 7.777 -1.231)')   # noqa: E501
CONTINUUM_COMPONENTS = katpoint.Catalogue([
    '3C286, radec, 13:31:08.3, 30:30:33, (400 50e3 0.1823 1.4757 -0.4739 0.0336 0 0 1 0.039 0.087 0)',          # noqa: E501
    '3C123, radec, 04:37:04.3753, 29:40:13.819, (400 50e3 3.1718 -0.1076 -0.1157)'
])


class TestLoaderKatdal:
    def setup(self):
        self.tempdir = tempfile.TemporaryDirectory()
        self.store = NpyFileChunkStore(self.tempdir.name)
        self.shape = (12, 96, len(ANTENNAS) * (len(ANTENNAS) + 1) * 2)
        self.telstate = katsdptelstate.TelescopeState()
        self._populate_telstate(self.telstate)

        self._open_patcher = mock.patch('katdal.open', autospec=True, side_effect=self._katdal_open)
        self._open_patcher.start()

    def _katdal_open(self, filename, **kwargs):
        """Mock implementation of katdal.open."""
        data_source = TelstateDataSource(
            self.view, self.cbid, self.stream_name, chunk_store=self.store, **kwargs)
        return VisibilityDataV4(data_source, **kwargs)

    @staticmethod
    def _populate_telstate(telstate):
        # Put in the minimum necessary for katdal to open the data set. This
        # may need to be updated in future as katdal evolves.
        telstate['sub_pool_resources'] = 'cbf_1,sdp_1,m000,m001,m002,m003'
        telstate['sub_product'] = 'c856M1k'
        telstate['sub_band'] = 'l'
        telstate['obs_params'] = {'observer': 'John Smith'}
        for i, ant in enumerate(ANTENNAS):
            telstate[f'{ant.name}_observer'] = ant.description
            telstate.add(f'{ant.name}_target', TARGET.description, 0.0)
            telstate.add(f'{ant.name}_activity', 'track', 1.0)
            # Point antennas in substantially different directions to test
            # parallactic angles.
            telstate.add(f'{ant.name}_pos_actual_scan_azim', np.deg2rad(i * 30.0))
            telstate.add(f'{ant.name}_pos_actual_scan_elev', np.deg2rad(30.0))
        telstate.add('obs_activity', 'track', 0.0)
        telstate.add('cbf_target', TARGET.description, 0.0)

    def _make_fake_dataset(self, **kwargs):
        chunks = (4, 8, self.shape[2])
        chunk_overrides = {
            'correlator_data': chunks,
            'flags': chunks,
            'weights': chunks,
            'weights_channel': chunks[:2]
        }
        default_kwargs = {
            'l0_shape': self.shape,
            'l0_chunk_overrides': chunk_overrides,
            'l1_flags_chunk_overrides': chunk_overrides
        }
        default_kwargs.update(kwargs)
        self.view, self.cbid, self.stream_name, _, _ = make_fake_data_source(
            self.telstate, self.store, **default_kwargs)

        # Link to RFI mask and band mask
        s_view = self.telstate.view(self.stream_name)
        bcp_view = self.telstate.view('bcp')
        acv_view = self.telstate.view('acv')
        self.telstate['sdp_model_base_url'] = 'http://models.invalid/'
        rfi_mask_key = self.telstate.join('model', 'rfi_mask', 'config')
        self.telstate[rfi_mask_key] = 'rfi_mask/config/unit_test.alias'
        s_view['src_streams'] = ['bcp']
        bcp_view['src_streams'] = ['acv']
        band_mask_key = self.telstate.join('model', 'band_mask', 'config')
        acv_view[band_mask_key] = 'band_mask/config/unit_test.alias'

        # Create a sky model
        sdp_archived_streams = self.telstate['sdp_archived_streams']
        sdp_archived_streams.append('continuum_image')
        self.telstate.delete('sdp_archived_streams')
        self.telstate['sdp_archived_streams'] = sdp_archived_streams
        continuum_view = self.telstate.view('continuum_image')
        continuum_view['stream_type'] = 'sdp.continuum_image'
        cs_view = self.telstate.view(self.telstate.join(self.cbid, 'continuum_image'))
        cs_view['targets'] = {TARGET.description: 'PKS_1934_63'}
        key = self.telstate.join('PKS_1934_63', 'target0', 'clean_components')
        cs_view[key] = {
            'description': TARGET.description,
            'components': [target.description for target in CONTINUUM_COMPONENTS]
        }

    def teardown(self):
        self._open_patcher.stop()
        self.tempdir.cleanup()

    def test_properties(self):
        # Pass a non-zero start channel to check that frequency-related
        # properties are not affected (it is only supposed to be a hint
        # rather than affecting indexing).
        self._make_fake_dataset()
        loader = LoaderKatdal('file:///fake_filename.rdb', {}, 24, None)
        np.testing.assert_array_equal(
            loader.antenna_diameters().to_value(u.m), [13.5] * len(ANTENNAS))
        assert_equal(loader.num_channels(), self.shape[1])
        assert_equal(loader.frequency(48), 1284 * u.MHz)
        assert_equal(loader.frequency(72), 1498 * u.MHz)
        assert_equal(loader.band(), 'L')
        assert_equal(loader.phase_centre()[0], Angle('19:39:25.03 hours').to(u.rad))
        assert_equal(loader.phase_centre()[1], Angle('-63:42:45.7 degrees').to(u.rad))
        assert_true(loader.has_feed_angles())
        assert_equal(
            loader.polarizations(),
            [
                polarization.STOKES_XX,
                polarization.STOKES_XY,
                polarization.STOKES_YX,
                polarization.STOKES_YY
            ]
        )
        assert_equal(loader.weight_scale(), np.sqrt(0.5))
        for i in range(self.shape[1]):
            assert_true(loader.channel_enabled(i))
        np.testing.assert_allclose(loader.longest_baseline(), 74.2 * u.m, rtol=0.001)
        assert_is_instance(loader.raw_data, VisibilityDataV4)
        assert_equal(loader.raw_target, TARGET)
        # Just a basic smoke test - doesn't test all the extra headers
        headers = loader.extra_fits_headers()
        assert_equal(headers['OBJECT'], 'PKS 1934-63')
        assert_equal(headers['INSTRUME'], 'c856M1k')
        assert_equal(headers['ONTIME'], 24.0)

    def test_sky_model(self):
        self._make_fake_dataset()
        loader = LoaderKatdal('file:///fake_filename.rdb', {}, 24, None)
        sky_model = loader.sky_model()
        assert_equal(sky_model._catalogue, CONTINUUM_COMPONENTS)

    def _collect(self, iterator):
        """Collect all the data from the ``data_iter`` iterator.

        Returns
        -------
        vis, weights, uvw, feed_angle1, feed_angle2 : np.ndarray
            Arrays of data
        chunks : Tuple[int]
            Number of rows in each loaded chunk
        """
        progress = 0
        chunks = []
        bls = len(ANTENNAS) * (len(ANTENNAS) - 1) // 2    # Excludes auto-correlations
        data = defaultdict(list)
        for chunk in iterator:
            assert_equal(chunk['total'], self.shape[0])
            assert_equal(chunk['progress'], progress + chunk['uvw'].shape[0] // bls)
            progress = chunk['progress']
            for field in ['uvw', 'weights', 'vis', 'feed_angle1', 'feed_angle2']:
                data[field].append(chunk[field])
            N = chunk['vis'].shape[1]
            # Check that nothing got accidentally upgraded to double precision
            assert_equal(chunk['vis'].dtype, np.complex64)
            assert_equal(chunk['weights'].dtype, np.float32)
            assert_equal(chunk['uvw'].dtype, np.float32)
            assert_equal(chunk['feed_angle1'].dtype, np.float32)
            assert_equal(chunk['feed_angle2'].dtype, np.float32)
            # Check that each array has the same number of rows
            assert_equal(chunk['uvw'].shape, (N, 3))
            assert_equal(chunk['weights'].shape, chunk['vis'].shape)
            assert_equal(chunk['feed_angle1'].shape, (N,))
            assert_equal(chunk['feed_angle2'].shape, (N,))
            chunks.append(N)
        assert_equal(progress, self.shape[0])

        # Concatenate the chunks back together
        vis = np.concatenate(data['vis'], axis=1)
        weights = np.concatenate(data['weights'], axis=1)
        uvw = np.concatenate(data['uvw'], axis=0)
        feed_angle1 = np.concatenate(data['feed_angle1'], axis=0)
        feed_angle2 = np.concatenate(data['feed_angle2'], axis=0)
        return vis, weights, uvw, feed_angle1, feed_angle2, tuple(chunks)

    def _get_expected(self, dataset, channel_mask=None):
        """Get values from a dataset.

        It must already have had the appropriate selection applied.
        """
        vis = dataset.vis[:]
        flags = dataset.flags[:]
        if channel_mask is not None:
            flags |= channel_mask[np.newaxis, :, np.newaxis]
        weights = dataset.weights[:] * np.logical_not(flags)
        # negate because katdal convention is ant2 - ant1, while katsdpimager
        # convention is the opposite.
        uvw = -np.stack([dataset.u, dataset.v, dataset.w], axis=2) * u.m
        pol_map = {
            'hh': polarization.STOKES_XX,
            'hv': polarization.STOKES_XY,
            'vh': polarization.STOKES_YX,
            'vv': polarization.STOKES_YY
        }
        # See comment in loader_katdal.py for explanation of pi/2 shift
        pa = np.deg2rad(dataset.parangle) - np.pi / 2
        pa = {ant.name: pa[:, i] for i, ant in enumerate(dataset.ants)}
        feed_angle1 = np.stack(
            [pa[prod[0][:-1]] for prod in dataset.corr_products],
            axis=1
        )
        feed_angle2 = np.stack(
            [pa[prod[1][:-1]] for prod in dataset.corr_products],
            axis=1
        )
        # Expected polarisation product per baseline
        polarizations = [
            pol_map[prod[0][-1:] + prod[1][-1:]]
            for prod in dataset.corr_products
        ]
        return vis, weights, uvw, feed_angle1, feed_angle2, polarizations

    def _test_data(self, *, options=(), channel0=36, channel1=72, max_chunk_vis=None,
                   missing=set(), channel_mask=None):
        loader = LoaderKatdal('file:///fake_filename.rdb', options, 24, None)
        bls = len(ANTENNAS) * (len(ANTENNAS) - 1) // 2    # Excludes auto-correlations
        vis, weights, uvw, feed_angle1, feed_angle2, chunks = \
            self._collect(loader.data_iter(channel0, channel1, max_chunk_vis=max_chunk_vis))

        dataset = katdal.open('file:///fake_filename.rdb')
        dataset.select(corrprods='cross', channels=np.s_[36:72], scans='track')
        if channel_mask is not None:
            channel_mask = channel_mask[channel0:channel1]
        e_vis, e_weights, e_uvw, e_feed_angle1, e_feed_angle2, e_polarizations = \
            self._get_expected(dataset, channel_mask)

        # Check the shapes before trying to shuffle data.
        N = self.shape[0] * bls
        C = e_vis.shape[1]
        P = 4
        assert_equal(vis.shape, (C, N, P))
        assert_equal(weights.shape, (C, N, P))
        assert_equal(uvw.shape, (N, 3))
        assert_equal(feed_angle1.shape, (N,))
        assert_equal(feed_angle2.shape, (N,))

        # The expected visibilities are all unique, so we can use them to
        # determine the permutation applied. We map each complex visibility to
        # its time-frequency-baseline coordinates.
        vis_map = {value: index for index, value in np.ndenumerate(e_vis)}
        for channel in range(C):
            for i in range(N):
                # Handling for missing correlation products assumes that the
                # first polarization product of a baseline will always be
                # present.
                baseline = dataset.corr_products[vis_map[vis[channel, i, 0]][2]]
                ant1 = baseline[0][:-1]
                ant2 = baseline[1][:-1]
                for j in range(P):
                    key = (ant1, ant2, loader.polarizations()[j])
                    if key in missing:
                        # Doesn't matter what the values are; it just has to be flagged
                        assert_equal(weights[channel, i, j], 0.0)
                        continue
                    # pop ensures that we don't see the same visibility twice
                    idx = vis_map.pop(vis[channel, i, j])
                    assert_equal(idx[1], channel)
                    assert_equal(loader.polarizations()[j], e_polarizations[idx[2]])
                    assert_equal(vis[channel, i, j], e_vis[idx])
                    assert_equal(weights[channel, i, j], e_weights[idx])
                    np.testing.assert_allclose(
                        uvw[i].to_value(u.m),
                        e_uvw[idx[0], idx[2]].to_value(u.m), rtol=1e-7
                    )
                    # Won't match exactly because the loader converts to float32
                    np.testing.assert_allclose(
                        feed_angle1[i], e_feed_angle1[idx[0], idx[2]], rtol=1e-6)
                    np.testing.assert_allclose(
                        feed_angle2[i], e_feed_angle2[idx[0], idx[2]], rtol=1e-6)
        assert_equal(len(vis_map), 0)
        loader.close()
        return chunks

    def test_basic(self):
        """Basic smoke test - can load everything in one chunk."""
        self._make_fake_dataset()
        self._test_data()

    def test_chunked(self):
        """Load via multiple chunks."""
        self._make_fake_dataset()
        bls = len(ANTENNAS) * (len(ANTENNAS) - 1) // 2    # Excludes auto-correlations
        chunks = self._test_data(max_chunk_vis=7 * 36 * 4 * bls)
        # It should round down to 4 times per chunk to align with data chunking
        assert_equal(chunks, (4 * bls, 4 * bls, 4 * bls))

    def test_max_chunk_vis_small(self):
        self._make_fake_dataset()
        with assert_logs(logger='katsdpimager.loader_katdal', level=logging.WARNING) as cm:
            self._test_data(max_chunk_vis=1)
        assert_in('Chunk size is 4 dumps but only 1 loaded at a time', cm.records[0].message)

    def test_missing_corr_product(self):
        corr_products = []
        # Apart from deleting one, also use a weird ordering to test the permutation
        for pol1 in 'hv':
            for pol2 in 'vh':
                for i in range(len(ANTENNAS)):
                    for j in range(i + 1):
                        if (i, j, pol1, pol2) != (2, 0, 'h', 'v'):
                            corr_products.append((f'{ANTENNAS[i].name}{pol1}',
                                                  f'{ANTENNAS[j].name}{pol2}'))
        assert_equal(len(corr_products), self.shape[2] - 1)
        self.shape = (self.shape[0], self.shape[1], len(corr_products))
        self._make_fake_dataset(bls_ordering_override=corr_products)
        self._test_data(missing={('m002', 'm000', polarization.STOKES_XY)})

    def _get_model(self, url, model_class, *, lazy=False):
        """Mock implementation of Fetcher.get."""
        if url == 'http://models.invalid/rfi_mask/config/unit_test.alias':
            ranges = astropy.table.Table(
                [[1200e6], [1300e6], [1000.0]],
                names=('min_frequency', 'max_frequency', 'max_baseline'),
                dtype=(np.float64, np.float64, np.float64)
            )
            ranges['min_frequency'].unit = u.Hz
            ranges['max_frequency'].unit = u.Hz
            ranges['max_baseline'].unit = u.m
            return katsdpmodels.rfi_mask.RFIMaskRanges(ranges, False)
        elif url == 'http://models.invalid/band_mask/config/unit_test.alias':
            ranges = astropy.table.Table(
                [[0.624], [0.626]],
                names=('min_fraction', 'max_fraction'),
                dtype=(np.float64, np.float64)
            )
            return katsdpmodels.band_mask.BandMaskRanges(ranges)
        else:
            raise RuntimeError(f'Unexpected URL {url}')

    @mock.patch('katsdpmodels.fetch.requests.Fetcher', autospec=True)
    def test_rfi_band_mask(self, fetcher_mock):
        fetcher_mock.return_value.get.side_effect = self._get_model
        self._make_fake_dataset()
        mask = np.zeros(self.shape[1], np.bool_)
        mask[39:51] = True      # RFI mask
        mask[60] = True         # Bank mask
        self._test_data(options=['--rfi-mask=config'], channel_mask=mask)

    def test_match(self):
        assert_false(LoaderKatdal.match('foo.h5'))
        assert_true(LoaderKatdal.match('foo.rdb'))
        assert_false(LoaderKatdal.match('foo.ms'))
        assert_true(LoaderKatdal.match('http://example.invalid/foo/bar.rdb?query=params#fragment'))
        assert_false(LoaderKatdal.match('http://example.invalid/foo/bar.ms?query=params#fragment'))
        assert_false(LoaderKatdal.match('http://[1.2.3/unmatch-bracket.rdb?query=params'))

    def test_missing_credentials(self):
        self._make_fake_dataset()
        with assert_raises(ValueError):
            self._test_data(options=['--access-key=foo'])
        with assert_raises(ValueError):
            self._test_data(options=['--secret-key=foo'])

    def test_credentials(self):
        self._make_fake_dataset()
        options = ['--access-key=foo', '--secret-key=bar']
        loader = LoaderKatdal('file:///fake_filename.rdb', options, 24, None)
        assert_equal(katdal.open.mock_calls[0][2].get('credentials'), ('foo', 'bar'))
        assert_equal(loader.command_line_options(), [], 'Credentials not redacted')



#!/usr/bin/env python3

"""
Create a simulated Measurement Set for use in comparing a number of imagers
against each other.

The time range, frequencies and phase centre are hard-coded in the script,
while the local sky model (consisting of point sources with polarisation but
no spectral shape) is read from lsm.txt, and the antennas from
meerkat_antennas.txt.
"""

import os

import numpy as np
import pandas
import katpoint
import casacore.tables

import astropy.units as u
from astropy.coordinates import SkyCoord, Angle
from astropy.time import Time
import astropy.io.ascii

# RASCIL gets upset if it can't find its data directory, but doesn't actually
# need it for this simulation. Just give it a fake one.
if 'RASCIL_DATA' not in os.environ:
    os.environ['RASCIL_DATA'] = '/'

import rascil.data_models                   # noqa: E402
import rascil.processing_components         # noqa: E402


freqs = 856e6 + 214e6 * (np.arange(4) + 0.5)
start_time = Time('2014-01-01T15:00:00.0', format='isot', scale='utc')
time_step = 4.0 * u.s
times = start_time + time_step * (np.arange(3600) + 0.5)
phase_centre = SkyCoord(ra='3h30m00s', dec='-35d00m00s', frame='icrs')

lsm = astropy.io.ascii.read('lsm.txt')
comps = [
    rascil.data_models.Skycomponent(
        SkyCoord(
            ra=Angle(row['ra'], unit='hour'),
            dec=Angle(row['dec'], unit='deg'),
            frame='icrs'
        ),
        [1e9],       # Arbitrary frequency
        flux=[[row['I'], row['Q'], row['U'], row['V']]]
    )
    for row in lsm
]

antennas = []
with open('meerkat_antennas.txt') as f:
    for line in f:
        if line:
            antennas.append(katpoint.Antenna(line))
location = antennas[0].ref_location
config = rascil.data_models.Configuration(
    location=location,
    names=[ant.name for ant in antennas],
    mount=np.repeat('azel', len(antennas)),
    xyz=[list(ant.position_enu) for ant in antennas],
    vp_type=np.repeat('unknown', len(antennas)),
    diameter=[ant.diameter.to_value(u.m) for ant in antennas],
    name='MeerKAT'
)

# RASCIL doesn't have very accurate UVW calculations
# (see https://gitlab.com/ska-telescope/rascil/-/issues/43), so we use
# katpoint instead.
baselines = pandas.MultiIndex.from_arrays(
    np.triu_indices(len(antennas)),
    names=('antenna1', 'antenna2')
)
ref_ant = antennas[0].array_reference_antenna()
target = katpoint.Target(phase_centre)
uvw_ant = target.uvw(antennas, times, ref_ant)
uvw = (uvw_ant[..., baselines.get_level_values('antenna1')]
       - uvw_ant[..., baselines.get_level_values('antenna2')])

vis_shape = (len(times), len(baselines), len(freqs), 4)
times_mjds = times.to_value('mjd') * 86400.0
bvis = rascil.data_models.BlockVisibility(
    frequency=freqs,
    channel_bandwidth=np.repeat(freqs[1] - freqs[0], len(freqs)),
    phasecentre=phase_centre,
    configuration=config,
    uvw=np.moveaxis(uvw, 0, -1),     # Move u/v/w axis to the end
    time=times_mjds,
    vis=np.zeros(vis_shape, np.complex64),
    integration_time=np.repeat(time_step.to_value(u.s), len(times)),
    flags=np.zeros(vis_shape, np.int_),
    baselines=baselines,
    polarisation_frame=rascil.data_models.PolarisationFrame('linear')
)
# Run the simulation
bvis = rascil.processing_components.dft_skycomponent_visibility(bvis, comps)
# Write the output
rascil.processing_components.export_blockvisibility_to_ms('simple.ms', [bvis])

# Make some fixes to the tables. See
# - https://gitlab.com/ska-telescope/rascil/-/issues/42
# - https://gitlab.com/ska-telescope/rascil/-/issues/44
with casacore.tables.table('simple.ms', readonly=False, ack=False) as t:
    timestamps = np.repeat(times_mjds, len(baselines))
    t.putcol('TIME', timestamps)
    t.putcol('TIME_CENTROID', timestamps)
with casacore.tables.table('simple.ms::ANTENNA', readonly=False, ack=False) as t:
    xyz = [u.Quantity(ant.location.geocentric).to_value(u.m) for ant in antennas]
    t.putcol('POSITION', np.array(xyz))
    t.putcol('MOUNT', ['ALT-AZ'] * len(antennas))



"""Python bindings for NVIDIA Tool Extensions.

This currently only handles a small subset of the API.

If the NVTX v3 headers were not detected at compile time, this module can still
be used but it will consist of no-ops.
"""

try:
    from ._nvtx_real import (             # noqa: F401
        initialize, RegisteredString, Domain, register_string, thread_range
    )
except ModuleNotFoundError:
    from ._nvtx_fallback import (         # noqa: F401
        initialize, RegisteredString, Domain, register_string, thread_range
    )



"""Tests for :py:mod:`katsdpimager.profiling`."""

import contextvars
import functools
import inspect
import io
from unittest import mock
from typing import Generator

from nose.tools import (
    assert_equal, assert_not_equal, assert_greater_equal,
    assert_is, assert_is_none, assert_is_not
)
import katsdpsigproc.abc
from katsdpsigproc.test.test_accel import device_test

from ..profiling import (
    Frame, Record, DeviceRecord, Profiler, CollectProfiler, FlamegraphProfiler,
    profile, profile_function, profile_generator
)


class TestFrame:
    def setup(self) -> None:
        self.frame1 = Frame('foo', {'x': 1, 'y': 2})
        self.frame2 = Frame('bar', {'x': 3, 'z': 'hello'}, self.frame1)

    def test_properties(self) -> None:
        assert_equal(self.frame1.name, 'foo')
        assert_equal(self.frame1.labels, {'x': 1, 'y': 2})
        assert_is_none(self.frame1.parent)
        assert_is(self.frame2.parent, self.frame1)

    def test_reuse(self) -> None:
        frame = Frame('bar', {'x': 3, 'z': 'hello'}, self.frame1)
        assert_is(frame, self.frame2)
        frame = Frame('bar', {'x': 3, 'z': 'goodbye'}, self.frame1)
        assert_is_not(frame, self.frame2)
        assert_equal(frame.labels, {'x': 3, 'z': 'goodbye'})

    def test_eq_hash(self) -> None:
        # Separate tree with identical properties
        frame1 = Frame('foo', {'x': 1, 'y': 2})
        frame2 = Frame('bar', {'x': 3, 'z': 'hello'}, frame1)
        assert_equal(frame1, self.frame1)
        assert_equal(frame2, self.frame2)
        assert_equal(hash(frame1), hash(self.frame1))
        assert_equal(hash(frame2), hash(self.frame2))
        # Same properties as frame2, but no parent
        frame = Frame('bar', {'x': 3, 'z': 'hello'})
        assert_not_equal(frame, frame2)
        assert_not_equal(hash(frame), hash(frame2))
        # Different properties, same parent
        frame = Frame('bar', {}, frame1)
        assert_not_equal(frame, frame2)
        assert_not_equal(hash(frame), hash(frame2))

    def test_stack(self) -> None:
        assert_equal(list(self.frame2.stack()), [self.frame1, self.frame2])

    def test_all_labels(self) -> None:
        assert_equal(self.frame1.all_labels(), {'x': 1, 'y': 2})
        assert_equal(self.frame2.all_labels(), {'x': 3, 'y': 2, 'z': 'hello'})


class TestRecord:
    def setup(self) -> None:
        self.frame = Frame('foo', {'str': 'hello', 'int': 1, 'float': 2.5})
        self.record = Record(self.frame, 12.5, 13.75)

    def test_construct(self) -> None:
        assert_equal(self.record.frame, self.frame)
        assert_equal(self.record.start_time, 12.5)
        assert_equal(self.record.stop_time, 13.75)

    def test_elapsed(self) -> None:
        assert_equal(self.record.elapsed, 1.25)

    def test_eq_hash(self) -> None:
        # Hash != comparisons could fail by chance, but it's very unlikely
        other_frame = Frame('bar', {})
        other = Record(other_frame, self.record.start_time, self.record.stop_time)
        assert_not_equal(self.record, other)
        assert_not_equal(hash(self.record), hash(other))
        other = Record(self.record.frame, self.record.start_time, self.record.stop_time)
        assert_equal(self.record, other)
        assert_equal(hash(self.record), hash(other))


def temporary_context(func):
    """Run a function with its own context."""
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        context = contextvars.copy_context()
        return context.run(func, *args, **kwargs)

    return wrapper


def use_self_profiler(func):
    """Modify a class or member to use self.profiler in all tests."""
    if inspect.isclass(func):
        for name, member in func.__dict__.items():
            if name.startswith('test_') and inspect.isfunction(member):
                setattr(func, name, use_self_profiler(member))
        return func

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        Profiler.set_profiler(self.profiler)
        return func(self, *args, **kwargs)

    return temporary_context(wrapper)


@profile_function()
def empty():
    pass


@mock.patch('time.monotonic', return_value=0.0)
@use_self_profiler
class TestProfiler:
    def setup(self) -> None:
        self.profiler = CollectProfiler()

    def test_simple(self, monotonic) -> None:
        with self.profiler.profile('foo'):
            monotonic.return_value = 1.0
        assert_equal(self.profiler.records, [Record(Frame('foo'), 0.0, 1.0)])

    def test_nested(self, monotonic) -> None:
        with self.profiler.profile('foo', {'x': 1}):
            monotonic.return_value += 1.0
            with self.profiler.profile('bar', {'y': 2}):
                monotonic.return_value += 1.0
            monotonic.return_value += 1.0
        frame1 = Frame('foo', {'x': 1})
        frame2 = Frame('bar', {'y': 2}, frame1)
        assert_equal(self.profiler.records, [
            Record(frame2, 1.0, 2.0),
            Record(frame1, 0.0, 3.0)
        ])

    @device_test
    def test_profile_device(self, monotonic, ctx: katsdpsigproc.abc.AbstractContext,
                            queue: katsdpsigproc.abc.AbstractCommandQueue) -> None:
        with self.profiler.profile('foo', {'x': 1}):
            with self.profiler.profile_device(queue, 'bar', {'y': 2}):
                pass
        assert_equal(len(self.profiler.device_records), 1)
        assert_equal(
            self.profiler.device_records[0].frame,
            Frame('bar', {'y': 2}, Frame('foo', {'x': 1}))
        )
        assert_greater_equal(self.profiler.device_records[0].elapsed, 0.0)

    def test_pause(self, monotonic) -> None:
        with self.profiler.profile('foo', {'x': 1}) as stopwatch:
            monotonic.return_value += 1.0
            stopwatch.stop()
            monotonic.return_value += 1.0
            stopwatch.start()
            monotonic.return_value += 1.0
        frame = Frame('foo', {'x': 1})
        assert_equal(self.profiler.records, [
            Record(frame, 0.0, 1.0),
            Record(frame, 2.0, 3.0)
        ])

    def test_context(self, monotonic) -> None:
        other_profiler = CollectProfiler()

        def inner():
            Profiler.set_profiler(other_profiler)
            with profile('foo', {'x': 1}):
                monotonic.return_value += 1.0

        context = contextvars.copy_context()
        context.run(inner)
        assert_is_not(Profiler.get_profiler(), other_profiler)
        frame = Frame('foo', {'x': 1})
        assert_equal(other_profiler.records, [Record(frame, 0.0, 1.0)])

    def test_profile_function_sequence_labels(self, monotonic):
        @profile_function('blah', ('x', 'y'))
        def inner(y, z, x=4):
            pass

        inner(2, 3)
        inner(y=5, z=6, x=7)
        assert_equal(self.profiler.records, [
            Record(Frame('blah', {'x': 4, 'y': 2}), 0.0, 0.0),
            Record(Frame('blah', {'x': 7, 'y': 5}), 0.0, 0.0)
        ])

    def test_profile_function_mapping_labels(self, monotonic):
        @profile_function('blah',
                          {'x': 'xx', 'y': lambda bound_args: 2 * bound_args.arguments['y']})
        def inner(xx, y):
            pass

        inner(2, 3)
        assert_equal(self.profiler.records, [
            Record(Frame('blah', {'x': 2, 'y': 6}), 0.0, 0.0)
        ])

    def test_profile_function_auto_name(self, monotonic):
        empty()
        assert_equal(self.profiler.records, [
            Record(Frame('katsdpimager.test.test_profiling.empty', {}), 0.0, 0.0)
        ])

    def test_profile_generator(self, monotonic):
        @profile_generator(name='slow_range', labels=('reps',))
        def slow_range(reps: int) -> Generator[int, None, None]:
            for i in range(reps):
                monotonic.return_value += 1.0
                yield i
            monotonic.return_value += 0.5

        result = []
        for item in slow_range(3):
            with profile('other_work'):
                monotonic.return_value += 2.0
                result.append(item)

        assert_equal(result, [0, 1, 2])
        frame = Frame('slow_range', {'reps': 3})
        frame2 = Frame('other_work')
        assert_equal(self.profiler.records, [
            Record(frame, 0.0, 1.0),
            Record(frame2, 1.0, 3.0),
            Record(frame, 3.0, 4.0),
            Record(frame2, 4.0, 6.0),
            Record(frame, 6.0, 7.0),
            Record(frame2, 7.0, 9.0),
            Record(frame, 9.0, 9.5)
        ])


class DummyEvent(katsdpsigproc.abc.AbstractEvent):
    def __init__(self, timestamp: float) -> None:
        self._timestamp = timestamp

    def wait(self) -> None:
        pass

    def time_since(self, prior_event: 'DummyEvent') -> float:
        return self._timestamp - prior_event._timestamp

    def time_till(self, next_event: 'DummyEvent') -> float:
        return next_event._timestamp - self._timestamp


class TestFlamegraphProfiler:
    def setup(self) -> None:
        self.profiler = FlamegraphProfiler()
        self.parent_frame = Frame('parent', {})
        self.child1_frame = Frame('child1', {}, self.parent_frame)
        self.child2_frame = Frame('child2', {}, self.parent_frame)

    def test_host(self) -> None:
        self.profiler.add_record(Record(self.child1_frame, 2.0, 3.0))
        self.profiler.add_record(Record(self.child2_frame, 4.0, 4.5))
        self.profiler.add_record(Record(self.parent_frame, 1.0, 11.0))
        out = io.StringIO()
        self.profiler.write_flamegraph(out)
        assert_equal(
            sorted(out.getvalue().splitlines()),
            [
                'parent 8500000',
                'parent;child1 1000000',
                'parent;child2 500000'
            ]
        )

    def test_device(self) -> None:
        self.profiler.add_device_record(
            DeviceRecord(self.child1_frame, DummyEvent(1.0), DummyEvent(3.0))
        )
        self.profiler.add_device_record(
            DeviceRecord(self.child2_frame, DummyEvent(5.0), DummyEvent(5.5))
        )
        out = io.StringIO()
        self.profiler.write_device_flamegraph(out)
        assert_equal(
            sorted(out.getvalue().splitlines()),
            [
                'parent;child1 2000000',
                'parent;child2 500000'
            ]
        )



"""Quality report for MeerKAT pipeline."""

from abc import abstractmethod, ABC
import itertools
import io
import json
import math
import logging
import xml.sax.saxutils
from datetime import datetime, timezone
from typing import List, Dict, Tuple, Optional

import pkg_resources
import jinja2
import numpy as np
import katpoint
import katdal
import katsdptelstate
import astropy.units as u

import matplotlib
import matplotlib.figure
import matplotlib.patches

import bokeh.embed
import bokeh.palettes
import bokeh.plotting
import bokeh.model
import bokeh.models
import bokeh.resources

import katsdpimager.metadata


FREQUENCY_PLOT_UNIT = u.MHz
FLUX_DENSITY_PLOT_UNIT = u.Jy
PIXEL_PLOT_UNIT = u.Jy / u.beam
ANGLE_PLOT_UNIT = u.deg
PALETTE = bokeh.palettes.colorblind['Colorblind'][8]
SI_FORMAT = bokeh.models.CustomJSHover(code="""
    const threshold = [1e-9, 1e-6, 1e-3, 1e0, 1e3,  1e6,  1e9, 1e12];
    const scale =     [1e12,  1e9,  1e6, 1e3, 1e0, 1e-3, 1e-6, 1e-9, 1e-12];
    const suffix =    [ "p",  "n",  "µ", "m",  "",  "k",  "M",  "G", "T"];
    var i = 0;
    if (value != value)
        return "N/A ";
    while (i < threshold.length && Math.abs(value) >= threshold[i])
        i++;
    return sprintf(format, value * scale[i]) + " " + suffix[i];
    """)

logger = logging.getLogger()


class SEFDModel(ABC):
    """Abstract class for system-equivalent flux density models."""

    @abstractmethod
    def __call__(self, frequencies: u.Quantity, effective: bool = False) -> u.Quantity:
        """Interpolate to given frequencies.

        If `effective` is true, the correlator efficiency is combined
        with the result.

        Given frequencies outside the support of the model return NaN.
        """


class PolynomialSEFDModel(SEFDModel):
    """SEFD model that uses a polynomial for each polarization.

    Parameters
    ----------
    min_frequency, max_frequency
        Frequency range over which the model is valid. For frequencies
        outside this range, NaN will be returned.
    coeffs
        2D array of polynomial coefficients. Each row corresponds to a
        single polarization, with coefficients increasing in exponent.
        The units must be flux density units e.g. Jy.
    frequency_unit
        Unit to which frequencies are converted for evaluation of the
        polynomial.
    correlator_efficiency
        Loss in SNR due to quantization in the correlator.
    """

    def __init__(self,
                 min_frequency: u.Quantity, max_frequency: u.Quantity,
                 coeffs: u.Quantity, frequency_unit: u.Unit,
                 correlator_efficiency: float) -> None:
        self.min_frequency = min_frequency
        self.max_frequency = max_frequency
        self._coeffs = coeffs.copy()
        self._frequency_unit = frequency_unit
        self.correlator_efficiency = correlator_efficiency

    def __call__(self, frequencies: u.Quantity, effective: bool = False) -> u.Quantity:
        # Note: don't use to_value, because that can return a value, and we're
        # going to modify x.
        x = frequencies.to(self._frequency_unit).value
        x[(frequencies < self.min_frequency) | (frequencies > self.max_frequency)] = np.nan
        pol_sefd = np.polynomial.polynomial.polyval(
            x, self._coeffs.value.T, tensor=True) << self._coeffs.unit
        # Take quadratic mean of individual polarizations
        sefd = np.sqrt(np.mean(np.square(pol_sefd), axis=0))
        if effective:
            sefd /= self.correlator_efficiency
        return sefd


def meerkat_sefd_model(band: str) -> SEFDModel:
    """Look up an SEFD model for MeerKAT.

    Parameters
    ----------
    band
        Frequency band for which the model will be applicable. Note that the model
        will likely not cover the edges of the band.

    Raises
    ------
    ValueError
        If no model is available for the given band
    """
    if band == 'L':
        coeffs = [
            [2.08778760e+02, 1.08462392e+00, -1.24639611e-03, 4.00344294e-07],   # H
            [7.57838984e+02, -2.24205001e-01, -1.72161897e-04, 1.11118471e-07]   # V
        ] * u.Jy
        return PolynomialSEFDModel(900.0 * u.MHz, 1670 * u.MHz, coeffs, u.MHz, 0.96)
    elif band == 'UHF':
        coeffs = [
            [1.20011355e+03, -7.08771871e-01, -1.46789604e-03, 1.29596990e-06],  # H
            [2.06467503e+03, -4.16858417e+00, 3.16131140e-03, -7.31852890e-07]   # V
        ] * u.Jy
        return PolynomialSEFDModel(580 * u.MHz, 1015 * u.MHz, coeffs, u.MHz, 0.96)
    else:
        raise ValueError(f'No SEFD model for band {band}')


class SVGMangler(xml.sax.saxutils.XMLGenerator):
    """Rip out the XML header and width and height attributes from an SVG.

    This is used to massage the SVG written by matplotlib to make it suitable for embedding.
    """

    def startDocument(self):
        pass      # Suppress the <?xml> header

    def startElement(self, name, attrs):
        if name == 'svg':
            # attrs is a non-mutable mapping, so we can't just delete from it
            attrs2 = dict(item for item in attrs.items() if item[0] not in {'width', 'height'})
            super().startElement(name, attrs.__class__(attrs2))
        else:
            super().startElement(name, attrs)


class CommonStats:
    """Information extracted from the dataset, independent of target."""

    def __init__(self, dataset: katdal.DataSet, telstate: katsdptelstate.TelescopeState) -> None:
        self.output_channels: List[int] = telstate['output_channels']
        self.channels: int = len(dataset.channels)
        self.channel_width = dataset.channel_width * u.Hz
        self.frequencies: u.Quantity = dataset.freqs * u.Hz
        self.antennas = dataset.ants
        # XXX for now this only supports MeerKAT. MeerKAT+ will be heterogeneous.
        self.sefd: Optional[u.Quantity] = None
        band = dataset.spectral_windows[dataset.spw].band
        try:
            sefd_model = meerkat_sefd_model(band)
        except ValueError:
            logger.warning('No SEFD model for band %s', band)
        else:
            self.sefd = sefd_model(self.frequencies, effective=True)


class TargetStats:
    """Collect all the statistics for a single target."""

    def __init__(self, dataset: katdal.DataSet,
                 common: CommonStats, target: katpoint.Target) -> None:
        self.common = common
        self.target = target
        # Status string for each channel
        self.status: List[str] = ['masked'] * common.channels
        for channel in common.output_channels:
            self.status[channel] = 'failed'
        # Peak per channel (NaN where missing)
        self.peak: u.Quantity = [math.nan] * common.channels * (u.Jy / u.beam)
        # Total flux density per (polarization, channel) (NaN where missing)
        self.totals: u.Quantity = {
            pol: [math.nan] * common.channels * u.Jy
            for pol in 'IQUV'
        }
        # Noise per channel (NaN where missing)
        self.noise: u.Quantity = [math.nan] * common.channels * (u.Jy / u.beam)
        # Estimate noise from weights (NaN where missing)
        self.weights_noise: u.Quantity = [math.nan] * common.channels * (u.Jy / u.beam)
        # Increase in noise due to imaging weights
        self.normalized_noise = [math.nan] * common.channels
        self.plots: Dict[str, str] = {}     # Divs to insert for plots returned by make_plots
        self.uv_coverage = ''
        self.frequency_range = bokeh.models.Range1d(
            self.common.frequencies[0].to_value(FREQUENCY_PLOT_UNIT),
            self.common.frequencies[-1].to_value(FREQUENCY_PLOT_UNIT),
            bounds='auto'
        )
        self.channel_range = bokeh.models.Range1d(0, self.common.channels - 1, bounds='auto')
        self.time_on_target = katsdpimager.metadata.time_on_target(dataset, target)
        self.model_natural_noise: Optional[u.Quantity] = None
        if self.common.sefd is not None and len(self.common.antennas) > 1:
            n = len(self.common.antennas)
            # Correlator efficiency is already folded in to self.common.sefd
            denom = math.sqrt(2 * n * (n - 1) * self.time_on_target * self.common.channel_width)
            self.model_natural_noise = self.common.sefd / denom / u.beam

        mask = katsdpimager.metadata.target_mask(dataset, target)
        self.timestamps = dataset.timestamps[mask]
        self.time_range = bokeh.models.Range1d(
            datetime.fromtimestamp(self.timestamps[0] - 0.5 * dataset.dump_period, timezone.utc),
            datetime.fromtimestamp(self.timestamps[-1] + 0.5 * dataset.dump_period, timezone.utc)
        )
        # Find contiguous time intervals on target
        delta = np.diff(mask, prepend=0, append=0)
        starts = np.nonzero(delta == 1)[0]
        ends = np.nonzero(delta == -1)[0] - 1
        self.time_intervals = list(zip(dataset.timestamps[starts] - 0.5 * dataset.dump_period,
                                       dataset.timestamps[ends] + 0.5 * dataset.dump_period))

        self.array_ant = dataset.sensor['Antennas/array/antenna'][0]
        self.ants = dataset.ants
        self.elevation = target.azel(timestamp=self.timestamps, antenna=self.array_ant)[1] << u.rad
        self.parallactic_angle = target.parallactic_angle(
            timestamp=self.timestamps, antenna=self.array_ant) << u.rad

    @property
    def name(self) -> str:
        return self.target.name

    @property
    def description(self) -> str:
        return self.target.description

    def _add_channel_range(self, fig: bokeh.plotting.Figure) -> None:
        fig.extra_x_ranges = {'channel': self.channel_range}
        fig.add_layout(bokeh.models.LinearAxis(x_range_name='channel', axis_label='Channel'),
                       'above')

    @staticmethod
    def _line_with_markers(fig: bokeh.plotting.Figure, *args, **kwargs) -> None:
        fig.line(*args, **kwargs)
        if 'line_color' in kwargs:
            kwargs['fill_color'] = kwargs['line_color']
        fig.circle(*args, **kwargs)

    def make_channel_data_source(self) -> bokeh.models.ColumnDataSource:
        data = {
            'frequency': self.common.frequencies.to_value(FREQUENCY_PLOT_UNIT),
            'status': self.status,
            'weights_noise': self.weights_noise.to_value(PIXEL_PLOT_UNIT),
            'noise': self.noise.to_value(PIXEL_PLOT_UNIT),
            'peak': self.peak.to_value(PIXEL_PLOT_UNIT)
        }
        for pol, total in self.totals.items():
            data[f'total_{pol}'] = total.to_value(FLUX_DENSITY_PLOT_UNIT)
        if self.model_natural_noise is not None:
            model_noise = self.model_natural_noise * self.normalized_noise
            data['model_noise'] = model_noise.to_value(PIXEL_PLOT_UNIT)
        return bokeh.models.ColumnDataSource(data)

    def make_time_data_source(self) -> bokeh.models.ColumnDataSource:
        data = {
            'time': [datetime.fromtimestamp(ts, timezone.utc) for ts in self.timestamps],
            'elevation': self.elevation.to_value(ANGLE_PLOT_UNIT),
            'parallactic_angle': self.parallactic_angle.to_value(ANGLE_PLOT_UNIT)
        }
        return bokeh.models.ColumnDataSource(data)

    def make_plot_status(self, source: bokeh.models.ColumnDataSource) -> bokeh.model.Model:
        factors = ['masked', 'failed', 'no-data', 'complete']
        fig = bokeh.plotting.figure(
            height=200,
            x_axis_label=f'Frequency ({FREQUENCY_PLOT_UNIT})',
            y_axis_label='Status',
            x_range=self.frequency_range,
            y_range=bokeh.models.FactorRange(factors=factors,
                                             bounds='auto',
                                             min_interval=len(factors)),
            sizing_mode='stretch_width', toolbar_location='below'
        )
        fig.cross(x='frequency', y='status', source=source, color=PALETTE[0])
        self._add_channel_range(fig)
        return fig

    def make_plot_snr(self, source: bokeh.models.ColumnDataSource) -> bokeh.model.Model:
        fig = bokeh.plotting.figure(
            x_axis_label=f'Frequency ({FREQUENCY_PLOT_UNIT})',
            y_axis_label=f'Flux density per beam ({PIXEL_PLOT_UNIT})',
            x_range=self.frequency_range,
            y_axis_type='log',
            sizing_mode='stretch_width'
        )
        fig.add_tools(bokeh.models.HoverTool(
            tooltips=[
                ('Frequency', f'@frequency{{0.0000}} {FREQUENCY_PLOT_UNIT}'),
                ('Channel', '$index'),
                ('Peak', f'@peak{{%.3f}}{PIXEL_PLOT_UNIT}'),
                ('Noise', f'@noise{{%.3f}}{PIXEL_PLOT_UNIT}'),
                ('Predicted noise (weights)', f'@weights_noise{{%.3f}}{PIXEL_PLOT_UNIT}'),
                ('Predicted noise (model)', f'@model_noise{{%.3f}}{PIXEL_PLOT_UNIT}')
            ],
            formatters={
                '@peak': SI_FORMAT,
                '@noise': SI_FORMAT,
                '@weights_noise': SI_FORMAT,
                '@model_noise': SI_FORMAT
            }
        ))
        self._line_with_markers(
            fig, x='frequency', y='peak', source=source, name='peak',
            line_color=PALETTE[0], legend_label='Peak')
        self._line_with_markers(
            fig, x='frequency', y='noise', source=source, name='noise',
            line_color=PALETTE[1], legend_label='Noise')
        self._line_with_markers(
            fig, x='frequency', y='weights_noise', source=source, name='weights_noise',
            line_color=PALETTE[2], legend_label='Predicted noise (weights)')
        if self.model_natural_noise is not None:
            self._line_with_markers(
                fig, x='frequency', y='model_noise', source=source, name='model_noise',
                line_color=PALETTE[3], legend_label='Predicted noise (model)')
        self._add_channel_range(fig)
        return fig

    def make_plot_flux_density(self, source: bokeh.models.ColumnDataSource) -> bokeh.model.Model:
        fig = bokeh.plotting.figure(
            x_axis_label=f'Frequency ({FREQUENCY_PLOT_UNIT})',
            y_axis_label=f'Flux density ({FLUX_DENSITY_PLOT_UNIT})',
            x_range=self.frequency_range,
            sizing_mode='stretch_width'
        )
        fig.add_tools(bokeh.models.HoverTool(
            tooltips=[
                ('Frequency', f'@frequency{{0.0000}} {FREQUENCY_PLOT_UNIT}'),
                ('Channel', '$index')
            ] + [
                (pol, f'@total_{pol}{{%.3f}}{FLUX_DENSITY_PLOT_UNIT}')
                for pol in self.totals.keys()
            ],
            formatters={f'@total_{pol}': SI_FORMAT for pol in self.totals.keys()}
        ))
        for i, pol in enumerate(self.totals.keys()):
            self._line_with_markers(
                fig, x='frequency', y=f'total_{pol}', source=source, name=f'total_{pol}',
                line_color=PALETTE[i], legend_label=pol)
        self._add_channel_range(fig)
        return fig

    def make_plot_elevation(self, source: bokeh.models.ColumnDataSource) -> bokeh.model.Model:
        unit_str = ANGLE_PLOT_UNIT.to_string('unicode')  # deg -> °
        fig = bokeh.plotting.figure(
            x_axis_label='Time (UTC)', x_axis_type='datetime',
            y_axis_label=f'Elevation ({unit_str})',
            x_range=self.time_range,
            y_range=[0.0, (90 * u.deg).to_value(ANGLE_PLOT_UNIT)],
            sizing_mode='stretch_width'
        )
        fig.add_tools(bokeh.models.HoverTool(
            tooltips=[
                ('Time', '@time{%Y-%m-%d %H:%M:%S}'),
                ('Elevation', f'@elevation {unit_str}')
            ],
            formatters={
                '@time': 'datetime'
            }
        ))
        fig.circle(x='time', y='elevation', source=source, line_color=PALETTE[0])
        # TODO: plot LST on the opposite axis?
        return fig

    def make_plot_parallactic_angle(
            self, source: bokeh.models.ColumnDataSource) -> bokeh.model.Model:
        unit_str = ANGLE_PLOT_UNIT.to_string('unicode')  # deg -> °
        fig = bokeh.plotting.figure(
            x_axis_label='Time (UTC)', x_axis_type='datetime',
            y_axis_label=f'Parallactic angle ({unit_str})',
            x_range=self.time_range,
            y_range=[(-180 * u.deg).to_value(ANGLE_PLOT_UNIT),
                     (180 * u.deg).to_value(ANGLE_PLOT_UNIT)],
            sizing_mode='stretch_width'
        )
        fig.add_tools(bokeh.models.HoverTool(
            tooltips=[
                ('Time', '@time{%Y-%m-%d %H:%M:%S}'),
                ('Angle', f'@parallactic_angle {unit_str}')
            ],
            formatters={
                '@time': 'datetime'
            }
        ))
        fig.circle(x='time', y='parallactic_angle', source=source, line_color=PALETTE[0])
        # TODO: plot LST on the opposite axis?
        return fig

    def make_plots(self) -> Dict[str, bokeh.model.Model]:
        """Generate Bokeh figures for the plots."""
        channel_source = self.make_channel_data_source()
        time_source = self.make_time_data_source()
        return {
            'status': self.make_plot_status(channel_source),
            'snr': self.make_plot_snr(channel_source),
            'flux_density': self.make_plot_flux_density(channel_source),
            'elevation': self.make_plot_elevation(time_source),
            'parallactic_angle': self.make_plot_parallactic_angle(time_source)
        }

    def make_uv_coverage(self) -> None:
        """Generate SVG showing UV coverage and store it internally.

        This is implemented in matplotlib (generating an SVG) rather than
        Bokeh because Bokeh doesn't support elliptic arcs.
        """
        matplotlib.use('SVG')
        fig = matplotlib.figure.Figure(figsize=(14, 14))
        ax = fig.subplots()
        ax.set_aspect('equal')

        def split_interval(start, end):
            """Split up long intervals into shorter ones.

            This is to avoid any ambiguities about which direction arcs should
            go if there is a track longer than 12 hours.
            """
            if end - start < 11 * 3600:
                yield start, end
            else:
                mid = (start + end) / 2
                yield from split_interval(start, mid)
                yield from split_interval(mid, end)

        for start, end in itertools.chain.from_iterable(
                split_interval(s, e) for s, e in self.time_intervals):
            # Find the Earth's rotation axis in the UVW basis (it's not a
            # constant due to precession etc, but close enough over a single
            # track). We use an offset of an hour rather than end time to
            # avoid numeric stability issues for very short tracks.
            bases = self.target.uvw_basis([start, start + 3600.0], self.array_ant)
            B = bases[..., 1] - bases[..., 0]
            # The axis of rotation will have the same UVW coordinates at both times,
            # and hence is in the null-space of B - so should correspond to a
            # singular value of zero.
            U, S, VH = np.linalg.svd(B)
            assert S[2] <= 1e-4 <= S[1]
            axis = U[:, 2]    # UVW coordinates of rotation axis (could be North or South)
            # Projection of the rotation axis gives direction of semi-minor
            # axis of the ellipse, and the w coordinate determines the ratio
            # between the axis lengths.
            flatten = abs(axis[2])
            angle = np.arctan2(axis[1], axis[0])

            all_uvw = np.stack(self.target.uvw(self.ants, [start, end], self.array_ant), axis=-1)
            for i in range(len(self.ants)):
                for j in range(i + 1, len(self.ants)):
                    uvw0, uvw1 = all_uvw[:, j] - all_uvw[:, i]  # Start and end uvw
                    # Project onto rotation axis to get centre of the ellipse.
                    # A lot of these calculations could be vectorised, but
                    # in practice I think the matplotlib arc rendering will
                    # dominate any potential savings.
                    mid = np.dot(uvw0, axis) * axis
                    a = np.linalg.norm(uvw0 - mid)
                    b = flatten * a
                    angle0 = np.arctan2(uvw0[1] - mid[1], uvw0[0] - mid[0])
                    angle1 = np.arctan2(uvw1[1] - mid[1], uvw1[0] - mid[0])
                    adiff = (angle1 - angle0) % (2 * np.pi)
                    # Ensure we draw the arc the shorter way around. We've
                    # limited the track length so that this should always be
                    # the correct direction.
                    if adiff >= np.pi:
                        angle0, angle1 = angle1, angle0
                    # matplotlib interprets arc endpoints relative to rotated ellipse.
                    angle0 -= angle
                    angle1 -= angle
                    arc = matplotlib.patches.Arc(
                        (mid[0], mid[1]), 2 * b, 2 * a, np.rad2deg(angle),
                        np.rad2deg(angle0),
                        np.rad2deg(angle1))
                    ax.add_patch(arc)
                    # Mirror image
                    arc = matplotlib.patches.Arc(
                        (-mid[0], -mid[1]), 2 * b, 2 * a, np.rad2deg(angle),
                        np.rad2deg(angle0) + 180,
                        np.rad2deg(angle1) + 180)
                    ax.add_patch(arc)
                    # Uncomment to plot dots on top of the tracks to verify.
                    # u, v, w = self.target.uvw([self.ants[i], self.ants[j]],
                    #                           np.arange(start, end, 256.0),
                    #                           self.array_ant)
                    # u = u[:, 1] - u[:, 0]
                    # v = v[:, 1] - v[:, 0]
                    # ax.plot(u, v, 'g.', markersize=1)
                    # ax.plot(-u, -v, 'g.', markersize=1)

        # Adjust data limits to be square and centred
        data_lim = ax.dataLim
        lim = max(-data_lim.x0, data_lim.x1, -data_lim.y0, data_lim.y1)
        ax.set_xlim(-lim, lim)
        ax.set_ylim(-lim, lim)
        ax.set_xlabel('m')
        ax.set_ylabel('m')
        raw_svg = io.BytesIO()
        fig.savefig(raw_svg, format='svg', bbox_inches='tight')

        # Clean up the SVG for embedding into the HTML.
        raw_svg.seek(0)
        out_svg = io.StringIO()
        mangler = SVGMangler(out_svg, encoding='utf-8', short_empty_elements=True)
        xml.sax.parse(raw_svg, mangler)
        self.uv_coverage = out_svg.getvalue()


def get_stats(dataset: katdal.DataSet,
              telstate: katsdptelstate.TelescopeState) -> Tuple[CommonStats, List[TargetStats]]:
    common = CommonStats(dataset, telstate)
    stats = [TargetStats(dataset, common, katpoint.Target(target))
             for target in telstate.get('targets', {})]
    stats_lookup = {target.description: target for target in stats}
    for ((target_desc, channel), status) in telstate.get('status', {}).items():
        stats_lookup[target_desc].status[channel] = status
    for ((target_desc, channel), peak) in telstate.get('peak', {}).items():
        stats_lookup[target_desc].peak[channel] = peak * (u.Jy / u.beam)
    for ((target_desc, channel, pol), total) in telstate.get('total', {}).items():
        stats_lookup[target_desc].totals[pol][channel] = total * u.Jy
    for ((target_desc, channel), noise) in telstate.get('noise', {}).items():
        stats_lookup[target_desc].noise[channel] = noise * (u.Jy / u.beam)
    for ((target_desc, channel), noise) in telstate.get('weights_noise', {}).items():
        stats_lookup[target_desc].weights_noise[channel] = noise * (u.Jy / u.beam)
    for ((target_desc, channel), normalized_noise) in telstate.get('normalized_noise', {}).items():
        stats_lookup[target_desc].normalized_noise[channel] = normalized_noise
    return common, stats


def format_duration(duration: u.Quantity) -> str:
    seconds = int(round(duration.to_value(u.s)))
    return '{}:{:02}:{:02}'.format(seconds // 3600, seconds // 60 % 60, seconds % 60)


def write_report(common_stats: CommonStats, target_stats: List[TargetStats],
                 filename: str) -> None:
    env = jinja2.Environment(
        loader=jinja2.PackageLoader('katsdpimager'),
        autoescape=True,
        undefined=jinja2.StrictUndefined
    )
    env.filters['duration'] = format_duration

    uv_coverage = [target.make_uv_coverage() for target in target_stats]
    plots = [target.make_plots() for target in target_stats]
    # Flatten all plots into one list to pass to bokeh
    flat_plots: List[Dict[str, bokeh.model.Model]] = []
    for plot_dict in plots:
        flat_plots.extend(plot_dict.values())
    script, divs = bokeh.embed.components(flat_plots)
    # Distribute divs back to individual objects
    i = 0
    for (target, plot_dict) in zip(target_stats, plots):
        for name in plot_dict:
            target.plots[name] = divs[i]
            i += 1

    resources = bokeh.resources.INLINE.render()
    static = {}
    for f in pkg_resources.resource_listdir('katsdpimager', 'static'):
        content = pkg_resources.resource_string('katsdpimager', 'static/' + f)
        static[f] = content.decode('utf-8')
    context = {
        'common': common_stats,
        'targets': target_stats,
        'resources': resources,
        'script': script,
        'static': static,
        'uv_coverage': uv_coverage
    }
    template = env.get_template('report.html.j2')
    template.stream(context).dump(filename)


def write_metadata(dataset: katdal.DataSet,
                   common_stats: CommonStats,
                   target_stats: List[TargetStats],
                   filename: str) -> None:
    half_channel = 0.5 * common_stats.channel_width
    metadata = {
        'ProductType': {
            'ProductTypeName': 'MeerKATReductionProduct',
            'ReductionName': 'Spectral Imager Report'
        },
        **katsdpimager.metadata.make_metadata(
            dataset, [target.target for target in target_stats],
            len(common_stats.output_channels),
            'Spectral-line imaging report'
        ),
        'MinFreq': (common_stats.frequencies[0] - half_channel).to_value(u.Hz),
        'MaxFreq': (common_stats.frequencies[-1] + half_channel).to_value(u.Hz),
        'Run': 0
    }
    with open(filename, 'w') as f:
        json.dump(metadata, f, allow_nan=False, indent=2)



"""Utilities to wrap skcuda.fft for imaging purposes."""

import threading

import numpy as np
import pkg_resources
import skcuda.fft
import skcuda.cufft
from katsdpsigproc import accel

import katsdpimager.types
from .profiling import profile_device

#: Forward FFT
FFT_FORWARD = 0
#: Inverse FFT, without normalization factor of 1/N
FFT_INVERSE = 1


class FftshiftTemplate:
    """Operation template for the equivalent of :py:meth:`np.fft.fftshift` on
    the device, in-place. The last two dimensions are shifted, and these
    dimensions must have even size. Because the size is even, this operation
    is its own inverse.

    Parameters
    ----------
    context : :class:`katsdpsigproc.abc.AbstractContext`
        Context for which kernels will be compiled
    dtype : numpy dtype
        Data type being stored
    tuning : dict, optional
        Tuning parameters (currently unused)
    """
    def __init__(self, context, dtype, tuning=None):
        self.context = context
        self.dtype = np.dtype(dtype)
        # TODO: autotune
        self.wgsx = 16
        self.wgsy = 8
        self.program = accel.build(
            context, "imager_kernels/fftshift.mako",
            {
                'ctype': katsdpimager.types.dtype_to_ctype(dtype),
                'wgsx': self.wgsx,
                'wgsy': self.wgsy
            },
            extra_dirs=[pkg_resources.resource_filename(__name__, '')])

    def instantiate(self, *args, **kwargs):
        return Fftshift(self, *args, **kwargs)


class Fftshift(accel.Operation):
    """Instantiation of :py:class:`FftshiftTemplate`.

    .. rubric:: Slots

    **data**
        Input and output array, transformed in-place

    Parameters
    ----------
    template : :class:`FftshiftTemplate`
        Operation template
    command_queue : :class:`katsdpsigproc.abc.AbstractCommandQueue`
        Command queue for the operation
    shape : tuple of int
        Shape of the data.
    allocator : :class:`~katsdpsigproc.accel.AbstractAllocator`, optional
        Allocator used to allocate unbound slots

    Raises
    ------
    ValueError
        if the first two dimensions in `shape` are not even
    """
    def __init__(self, template, command_queue, shape, allocator=None):
        super().__init__(command_queue, allocator)
        self.template = template
        self.kernel = template.program.get_kernel('fftshift')
        if shape[-1] % 2 != 0 or shape[-2] % 2 != 0:
            raise ValueError('First two dimensions of shape must be even')
        self.slots['data'] = accel.IOSlot(shape, template.dtype)

    def _run(self):
        data = self.buffer('data')
        items_x = data.shape[-1] // 2
        items_y = data.shape[-2] // 2
        if len(data.shape) == 2:
            items_z = 1
        else:
            items_z = data.shape[0] * int(np.product(data.padded_shape[1:-2]))
        row_stride = data.padded_shape[-1]
        slice_stride = data.padded_shape[-2] * data.padded_shape[-1]
        with profile_device(self.command_queue, 'fftshift'):
            self.command_queue.enqueue_kernel(
                self.kernel,
                [
                    data.buffer,
                    np.int32(row_stride),
                    np.int32(slice_stride),
                    np.int32(items_x), np.int32(items_y),
                    np.int32(items_y * row_stride)
                ],
                global_size=(accel.roundup(items_x, self.template.wgsx),
                             accel.roundup(items_y, self.template.wgsy),
                             items_z),
                local_size=(self.template.wgsx, self.template.wgsy, 1)
            )


class FftTemplate:
    r"""Operation template for a forward or reverse FFT. The transformation is
    done over the last N dimensions, with the remaining dimensions for batching
    multiple arrays to be transformed. Dimensions before the first N must not
    be padded.

    This template bakes in more information than most (data shapes), which is
    due to constraints in CUFFT.

    The template can specify real->complex, complex->real, or
    complex->complex. In the last case, the same template can be used to
    instantiate forward or inverse transforms. Otherwise, real->complex
    transforms must be forward, and complex->real transforms must be inverse.

    For real<->complex transforms, the final dimension of the padded shape
    need only be :math:`\lfloor\frac{L}{2}\rfloor + 1`, where :math:`L` is the
    last element of `shape`.

    Parameters
    ----------
    context : :class:`katsdpsigproc.cuda.Context`
        Context for the operation
    N : int
        Number of dimensions for the transform
    shape : tuple
        Shape of the input (N or more dimensions)
    dtype_src : {`np.float32, `np.float64`, `np.complex64`, `np.complex128`}
        Data type for input
    dtype_dest : {`np.float32, `np.float64`, `np.complex64`, `np.complex128`}
        Data type for output
    padded_shape_src : tuple
        Padded shape of the input
    padded_shape_dest : tuple
        Padded shape of the output
    tuning : dict, optional
        Tuning parameters (currently unused)
    """
    def __init__(self, context, N, shape, dtype_src, dtype_dest,
                 padded_shape_src, padded_shape_dest, tuning=None):
        if len(padded_shape_src) != len(shape):
            raise ValueError('padded_shape_src and shape must have same length')
        if len(padded_shape_dest) != len(shape):
            raise ValueError('padded_shape_dest and shape must have same length')
        if padded_shape_src[:-N] != shape[:-N]:
            raise ValueError('Source must not be padded on batch dimensions')
        if padded_shape_dest[:-N] != shape[:-N]:
            raise ValueError('Destination must not be padded on batch dimensions')
        self.context = context
        self.shape = shape
        self.dtype_src = np.dtype(dtype_src)
        self.dtype_dest = np.dtype(dtype_dest)
        self.padded_shape_src = padded_shape_src
        self.padded_shape_dest = padded_shape_dest
        # CUDA 7.0 CUFFT has a bug where kernels are run in the default stream
        # instead of the requested one, if dimensions are up to 1920. There is
        # a patch, but there is no query to detect whether it has been
        # applied.
        self.needs_synchronize_workaround = (
            skcuda.cufft.cufftGetVersion() < 8000 and any(x <= 1920 for x in shape[:N])
        )
        batches = int(np.product(padded_shape_src[:-N]))
        with context:
            self.plan = skcuda.fft.Plan(
                shape[-N:], dtype_src, dtype_dest, batches,
                inembed=np.array(padded_shape_src[-N:], np.int32),
                istride=1,
                idist=int(np.product(padded_shape_src[-N:])),
                onembed=np.array(padded_shape_dest[-N:], np.int32),
                ostride=1,
                odist=int(np.product(padded_shape_dest[-N:])),
                auto_allocate=False)
        # The stream and work area are associated with the plan rather than
        # the execution, so we need to serialise executions.
        self.lock = threading.RLock()

    def instantiate(self, *args, **kwargs):
        return Fft(self, *args, **kwargs)


class Fft(accel.Operation):
    """Forward or inverse Fourier transformation.

    .. rubric:: Slots

    **src**
        Input data
    **dest**
        Output data
    **work_area**
        Scratch area for work. The contents should not be used; it is made
        available so that it can be aliased with other scratch areas.

    Parameters
    ----------
    command_queue : :class:`katsdpsigproc.cuda.CommandQueue`
        Command queue for the operation
    template : :class:`FftTemplate`
        Operation template
    mode : {:data:`FFT_FORWARD`, :data:`FFT_INVERSE`}
        FFT direction
    allocator : :class:`~katsdpsigproc.accel.AbstractAllocator`, optional
        Allocator used to allocate unbound slots
    """
    def __init__(self, template, command_queue, mode, allocator=None):
        super().__init__(command_queue, allocator)
        self.template = template
        src_shape = list(template.shape)
        dest_shape = list(template.shape)
        if template.dtype_src.kind != 'c':
            if mode != FFT_FORWARD:
                raise ValueError('R2C transform must use FFT_FORWARD')
            dest_shape[-1] = template.shape[-1] // 2 + 1
        if template.dtype_dest.kind != 'c':
            if mode != FFT_INVERSE:
                raise ValueError('C2R transform must use FFT_INVERSE')
            src_shape[-1] = template.shape[-1] // 2 + 1
        src_dims = [accel.Dimension(d[0], min_padded_size=d[1], exact=True)
                    for d in zip(src_shape, template.padded_shape_src)]
        dest_dims = [accel.Dimension(d[0], min_padded_size=d[1], exact=True)
                     for d in zip(dest_shape, template.padded_shape_dest)]
        self.slots['src'] = accel.IOSlot(src_dims, template.dtype_src)
        self.slots['dest'] = accel.IOSlot(dest_dims, template.dtype_dest)
        self.slots['work_area'] = accel.IOSlot((template.plan.worksize,), np.uint8)
        self.mode = mode

    def _run(self):
        src_buffer = self.buffer('src')
        dest_buffer = self.buffer('dest')
        work_area_buffer = self.buffer('work_area')
        context = self.command_queue.context
        with context, self.template.lock:
            skcuda.cufft.cufftSetStream(self.template.plan.handle,
                                        self.command_queue._pycuda_stream.handle)
            self.template.plan.set_work_area(work_area_buffer.buffer)
            if self.mode == FFT_FORWARD:
                with profile_device(self.command_queue, 'fft'):
                    skcuda.fft.fft(src_buffer.buffer, dest_buffer.buffer, self.template.plan)
            else:
                with profile_device(self.command_queue, 'ifft'):
                    skcuda.fft.ifft(src_buffer.buffer, dest_buffer.buffer, self.template.plan)
            if self.template.needs_synchronize_workaround:
                context._pycuda_context.synchronize()



#!/usr/bin/env python3
import argparse
import logging
import os
import io
import uuid
import shutil
import json
from contextlib import closing
import gc
import shlex

import numpy as np
from astropy import units
import katsdpservices
import katsdptelstate
import katdal
from katsdpsigproc import accel
import katsdpimageutils.render
import astropy.io.fits as fits
import astropy.units as u

from katsdpimager import frontend, loader, progress, metadata, arguments, profiling
from katsdpimager.profiling import profile, profile_function
import katsdpimager.io    # Use full name to avoid conflict with stdlib io


logger = logging.getLogger()


class Writer(frontend.Writer):
    def __init__(self, args, dataset):
        self.args = args
        self.uuid = uuid.uuid4()

        options = frontend.command_line_options(
            args, dataset, {'input_file', 'output_dir', 'prefix', 'stream', 'input_option'})
        self.extra_fits_headers = fits.Header()
        self.extra_fits_headers['HISTORY'] = \
            'Command line options: ' + ' '.join(shlex.quote(arg) for arg in options)

        raw_data = dataset.raw_data
        if not isinstance(raw_data, katdal.DataSet):
            raise RuntimeError('Only katdal data sets are supported')
        self.common_metadata = {
            'ProductType': {
                'ProductTypeName': 'FITSImageProduct',
                'ReductionName': 'Spectral Image'
            },
            **metadata.make_metadata(raw_data, [dataset.raw_target], 1,
                                     'Spectral-line image')
        }
        telstate = raw_data.source.telstate.root()
        namespace = telstate.join(raw_data.source.capture_block_id, args.stream)
        self.telstate = telstate.view(namespace)

    def channel_already_done(self, dataset, channel):
        sub_key = (dataset.raw_target.description, channel)
        return self.telstate.get_indexed('status', sub_key) is not None

    def needs_fits_image(self, name):
        return name == 'clean'

    def needs_fits_grid(self, name):
        return False

    @profile_function()
    def _downsample(self, image, headers):
        """Create a reduced HDU appropriate for the thumbnail.

        Only the first polarization is retained, and it is downsampled to
        approximately match the thumbnail image size. katsdpimageutils (via
        matplotlib) handles the final rescaling, so this is just to speed
        things up (probably at a small cost in accuracy, but the thumbnail
        image is just for a quick look, not science).
        """
        MIN_SIZE = 512
        scale = 1
        # Keep things simple by only considering exact scale factors, which
        # makes scaling down easy. is_smooth enforces multiples of 8,
        # so we can always go at least that far before encountering
        # divisibility problems.
        while (image.shape[-1] % (scale * 2) == 0
               and image.shape[-1] // (scale * 2) >= MIN_SIZE
               and image.shape[-2] % (scale * 2) == 0
               and image.shape[-2] // (scale * 2) >= MIN_SIZE):
            scale *= 2
        headers = headers.copy()
        headers['CDELT1'] *= scale
        headers['CDELT2'] *= scale
        # FITS interprets 1.0 as the centre of the first pixel
        headers['CRPIX1'] = (headers['CRPIX1'] - 0.5) / scale + 0.5
        headers['CRPIX2'] = (headers['CRPIX2'] - 0.5) / scale + 0.5

        image = image[:, :1, ...]    # Keep only the first polarization
        image = np.add.reduceat(image, range(0, image.shape[-1], scale), axis=-1)
        image = np.add.reduceat(image, range(0, image.shape[-2], scale), axis=-2)
        image /= scale * scale
        hdu = fits.PrimaryHDU(image, headers)
        hdu.update_header()    # Updates NAXIS*
        return hdu

    @profile_function(labels=['name', 'channel'])
    def write_fits_image(self, name, description, dataset, image, image_parameters, channel,
                         beam=None, bunit='Jy/beam'):
        if name != 'clean':
            return
        # Add a unique component to the directory name so that it will be
        # unique even if the task dies and is re-run.
        output_dir = '{}_{:05}_{}'.format(self.args.prefix, channel, self.uuid)
        output_dir = os.path.join(self.args.output_dir, output_dir)
        tmp_dir = output_dir + '.writing'
        base_filename = '{}_{:05}.fits'.format(self.args.prefix, channel)
        filename = os.path.join(tmp_dir, base_filename)
        freq = image_parameters.wavelength.to(units.Hz, equivalencies=units.spectral()).value
        channel_width = dataset.raw_data.channel_width
        metadata = {
            **self.common_metadata,
            'FITSImageFilename': [base_filename],
            'PNGThumbNailFileName': [base_filename + '.tnail.png'],
            'CenterFrequency': freq,
            'MinFreq': freq - 0.5 * channel_width,
            'MaxFreq': freq + 0.5 * channel_width,
            'Run': channel
        }
        os.mkdir(tmp_dir)
        try:
            with progress.step('Write {}'.format(description)):
                fits_image, fits_headers = katsdpimager.io.write_fits_image(
                    dataset, image, image_parameters, filename,
                    channel, beam, bunit, self.extra_fits_headers)
            with progress.step('Write thumbnail'):
                hdu = self._downsample(fits_image, fits_headers)
                with profile(
                    'katsdpimageutils.render.write_image',
                    labels={'filename': filename + '.tnail.png'}
                ):
                    katsdpimageutils.render.write_image(
                        hdu, filename + '.tnail.png',
                        width=650, height=500
                    )
            with open(os.path.join(tmp_dir, 'metadata.json'), 'w') as f:
                json.dump(metadata, f, allow_nan=False, indent=2)
            os.rename(tmp_dir, output_dir)
            # Free some things ahead of the gc.collect below
            del hdu
            del fits_image
            del fits_headers
        except Exception:
            # Make a best effort to clean up
            shutil.rmtree(tmp_dir, ignore_errors=True)
            raise
        finally:
            # Something in the PNG writing causes memory to not get freed,
            # and the garbage collector is not kicking in properly on its
            # own.
            with profile('gc.collect'):
                gc.collect()

    def write_fits_grid(self, name, description, fftshift, grid_data, image_parameters, channel):
        pass

    def skip_channel(self, dataset, image_parameters, channel):
        sub_key = (dataset.raw_target.description, channel)
        self.telstate.set_indexed('status', sub_key, 'no-data')

    def _set_statistic(self, key, sub_key, value):
        # The imaging adds floating-point values in a non-deterministic order
        # and hence produces slightly different results every time. In the
        # unlikely event that we filled in statistics in a previous run but
        # didn't make it to setting the completion flag, this can fail.
        try:
            self.telstate.set_indexed(key, sub_key, value)
        except katsdptelstate.ImmutableKeyError as exc:
            logger.warning("%s", exc)

    @profile_function()
    def statistics(self, dataset, channel, **kwargs):
        sub_key = (dataset.raw_target.description, channel)
        peak = kwargs['peak']
        if np.isfinite(peak):
            self._set_statistic('peak', sub_key, peak)
        for pol, total in kwargs['totals'].items():
            self._set_statistic('total', sub_key + (pol,), total)
        if kwargs.get('weights_noise') is not None:
            self._set_statistic('weights_noise', sub_key, kwargs['weights_noise'])
        for key in ['noise', 'normalized_noise', 'major', 'minor', 'psf_patch_size',
                    'compressed_vis']:
            self._set_statistic(key, sub_key, kwargs[key])

        image_p = kwargs['image_parameters']
        self._set_statistic('pixel_size', sub_key, float(image_p.pixel_size))
        self._set_statistic('pixels', sub_key, image_p.pixels)
        self._set_statistic('cell_size', sub_key, image_p.cell_size.to_value(u.m))

        grid_p = kwargs['grid_parameters']
        for key in ['image_oversample', 'kernel_width']:
            self._set_statistic(key, sub_key, getattr(grid_p.fixed, key))
        for key in ['w_slices', 'w_planes']:
            self._set_statistic(key, sub_key, getattr(grid_p, key))
        self._set_statistic('grid_oversample', sub_key, grid_p.fixed.oversample)

        # statistics() is the last step in process_channel, so if we get this
        # far, the channel is fully processed.
        self.telstate.set_indexed('status', sub_key, 'complete')

    def finalize(self, dataset, start_channel, stop_channel):
        sub_key = (dataset.raw_target.description, start_channel, stop_channel)
        profiler = profiling.Profiler.get_profiler()

        flamegraph = io.StringIO()
        profiler.write_flamegraph(flamegraph)
        self._set_statistic('flamegraph', sub_key, flamegraph.getvalue())

        device_flamegraph = io.StringIO()
        profiler.write_device_flamegraph(device_flamegraph)
        self._set_statistic('device_flamegraph', sub_key, device_flamegraph.getvalue())


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('input_file', type=str, metavar='INPUT',
                        help='Input data set')
    parser.add_argument('output_dir', type=str, metavar='DIR',
                        help='Parent directory for output')
    parser.add_argument('prefix', type=str,
                        help='Prefix for output directories and filenames')
    parser.add_argument('stream', type=str,
                        help='Stream name for telescope state outputs')
    parser.add_argument('--log-level', type=str, metavar='LEVEL',
                        help='Logging level [INFO]')

    frontend.add_options(parser)

    # Arguments only supported by offline imager.py
    parser.set_defaults(host=False, vis_limit=None)
    return parser


def main():
    parser = get_parser()
    args = parser.parse_args(namespace=arguments.SmartNamespace())
    katsdpservices.setup_logging()
    if args.log_level is not None:
        logger.setLevel(args.log_level.upper())

    profiling.Profiler.set_profiler(profiling.FlamegraphProfiler())

    with closing(loader.load(args.input_file, args.input_option,
                             args.start_channel, args.stop_channel)) as dataset:
        writer = Writer(args, dataset)
        context = accel.create_some_context(interactive=False, device_filter=lambda x: x.is_cuda)
        queue = context.create_command_queue()
        frontend.run(args, context, queue, dataset, writer)
        # frontend.run modifies args.stop_channel in place, so even if it
        # wasn't specified by the user it will now be valid.
        writer.finalize(dataset, args.start_channel, args.stop_channel)


if __name__ == '__main__':
    main()



"""Variants of standard maths functions written for speed."""

import numpy as np
import numba


@numba.vectorize([numba.complex64(numba.float32),
                  numba.complex128(numba.float64)])
def expj2pi(x):
    """Equivalent to ``expj(2 * np.pi * x)`` where `x` is real.

    x is reduced to a small value before multiplication, which
    improves precision at a small cost in performance.
    """
    y = 2 * np.pi * (x - np.rint(x))
    return complex(np.cos(y), np.sin(y))


def nansum(x, *args, **kwargs):
    """Like np.nansum, provided `x` is a floating-point type."""
    return np.sum(x, *args, where=~np.isnan(x), **kwargs)



import contextlib
from typing import Optional, Generator, Union

from ._nvtx import ffi, lib


def initialize() -> None:
    lib.nvtxInitialize(ffi.NULL)


class RegisteredString:
    def __init__(self, handle: object) -> None:
        self._handle = handle


class Domain:
    def __init__(self, name: str) -> None:
        self._handle = lib.nvtxDomainCreateW(name)

    def __del__(self, _nvtxDomainDestroy=lib.nvtxDomainDestroy) -> None:
        _nvtxDomainDestroy(self._handle)


def register_string(value: str, domain: Optional[Domain] = None) -> RegisteredString:
    domain_handle = domain._handle if domain is not None else ffi.NULL
    handle = lib.nvtxDomainRegisterStringW(domain_handle, value)
    return RegisteredString(handle)


@contextlib.contextmanager
def thread_range(name: Union[str, RegisteredString], *,
                 payload: Union[None, int, float] = None,
                 domain: Optional[Domain] = None) -> Generator[None, None, None]:
    event = ffi.new("nvtxEventAttributes_t *")
    event.version = lib.NVTX_VERSION
    event.size = lib.NVTX_EVENT_ATTRIB_STRUCT_SIZE
    if isinstance(name, RegisteredString):
        event.messageType = lib.NVTX_MESSAGE_TYPE_REGISTERED
        event.message.registered = name._handle
    else:
        c_name = ffi.new("wchar_t[]", name)  # Assign to a var to keep it alive
        event.messageType = lib.NVTX_MESSAGE_TYPE_UNICODE
        event.message.unicode = c_name
    if payload is not None:
        if isinstance(payload, int):
            event.payloadType = lib.NVTX_PAYLOAD_TYPE_INT64
            event.payload.llValue = payload
        elif isinstance(payload, float):
            event.payloadType = lib.NVTX_PAYLOAD_TYPE_DOUBLE
            event.payload.dValue = payload
        else:
            raise TypeError('Payload must be None, int or float')

    # Note: we don't check the return values for errors, because if no
    # profiling tool is active then these functions "fail".
    domain_handle = domain._handle if domain is not None else ffi.NULL
    lib.nvtxDomainRangePushEx(domain_handle, event)
    try:
        yield
    finally:
        lib.nvtxDomainRangePop(domain_handle)



"""Functions for recording profiling information."""

from abc import abstractmethod, ABC
import time
import collections.abc
from collections import deque
import contextlib
from contextvars import ContextVar, Token
import functools
import inspect
import io
import weakref
import typing
from typing import (
    List, Tuple, Dict, Deque, Mapping, Sequence, Callable, TextIO,
    Generator, Union, Optional, TypeVar, Any
)

import katsdpsigproc.abc

from . import nvtx


_T = TypeVar('_T')
_LabelSpec = Union[Sequence[str],
                   Mapping[str, Union[Callable[[inspect.BoundArguments], Any], str]]]
_domain = nvtx.Domain(__name__)


class Frame:
    """Stack frame.

    Where possible, identical frames are represented by the same object to save
    memory.

    Stack frames may have arbitrary key/value labels associated to assist with
    aggregation. Label values must be CSV-compatible e.g. strings, ints and
    floats. The following labels are standardised:

    - channel
    - start_channel, stop_channel
    """

    name: str
    labels: Dict[str, Any]
    parent: Optional['Frame']
    _name_handle: nvtx.RegisteredString
    _hash: int
    _children: weakref.WeakValueDictionary

    def __new__(cls, name, labels: Mapping[str, Any] = {},
                parent: Optional['Frame'] = None) -> 'Frame':
        labels = dict(labels)
        key = (name, tuple(sorted(labels.items())))
        if parent is not None:
            frame = parent._children.get(key)
            if frame is not None:
                return frame
        frame = super().__new__(cls)
        frame.name = name
        frame.labels = labels
        frame.parent = parent
        frame._name_handle = nvtx.register_string(name, domain=_domain)
        frame._hash = hash((key, parent))
        frame._children = weakref.WeakValueDictionary()
        if parent is not None:
            parent._children[key] = frame
        return frame

    def __eq__(self, other: object) -> bool:
        if type(other) != Frame:
            return NotImplemented
        assert isinstance(other, Frame)    # To keep mypy happy
        if self._hash != other._hash:
            return False
        return (self.name == other.name and self.labels == other.labels
                and self.parent == other.parent)

    def __hash__(self) -> int:
        return self._hash

    def stack(self) -> Generator['Frame', None, None]:
        """Iterate over the frames in the stack, from root to leaf."""
        if self.parent is not None:
            yield from self.parent.stack()
        yield self

    def all_labels(self) -> Mapping[str, Any]:
        """Collapse all labels in the stack to a single dictionary.

        Labels closer to the top of the stack update those below.
        """
        labels = {}
        for frame in self.stack():
            labels.update(frame.labels)
        return labels

    def __repr__(self) -> str:
        if self.parent is None:
            return f'Frame({self.name!r}, {self.labels!r})'
        else:
            return f'Frame({self.name!r}, {self.labels!r}, {self.parent!r})'


class Record:
    """A record of some process with a start and a stop time.

    The start and stop times are as returned by :func:`time.monotonic`.
    """

    def __init__(self, frame: Frame, start_time: float, stop_time: float) -> None:
        self.frame = frame
        self.start_time = start_time
        self.stop_time = stop_time

    @property
    def elapsed(self) -> float:
        return self.stop_time - self.start_time

    def __eq__(self, other: object) -> bool:
        if type(other) == Record:
            return vars(self) == vars(other)
        else:
            return NotImplemented

    def __hash__(self) -> int:
        return hash((self.frame, self.start_time, self.stop_time))

    def __repr__(self) -> str:
        return f'Record({self.frame!r}, {self.start_time!r}, {self.stop_time!r})'


class DeviceRecord:
    """A record of some on-device process with a start and a stop event."""

    def __init__(self, frame: Frame,
                 start_event: katsdpsigproc.abc.AbstractEvent,
                 stop_event: katsdpsigproc.abc.AbstractEvent) -> None:
        self.frame = frame
        self.start_event = start_event
        self.stop_event = stop_event

    @property
    def elapsed(self) -> float:
        """Get elapsed time.

        Note that this may block until the stop event completes.
        """
        return self.stop_event.time_since(self.start_event)


class Stopwatch(contextlib.ContextDecorator):
    """Measures intervals of time and add records of them to a profiler.

    Each interval is bracketed by calls to :meth:`start` and :meth:`stop`.
    It can also be (and is designed to be) used as a context manager, in
    which case it will start on entry and stop (if necessary) on exit.
    Within the context manager it is possible to use stop/start to pause
    the stopwatch.
    """

    def __init__(self, profiler: 'Profiler', frame: Frame):
        self._profiler = profiler
        self._frame = frame
        self._start_time: Optional[float] = None
        self._token: Optional[Token] = None
        self._nvtx_ctx = None

    def __enter__(self) -> 'Stopwatch':
        self.start()
        return self

    def __exit__(self, *args) -> None:
        if self._start_time is not None:
            self.stop()

    def start(self) -> None:
        if self._start_time is not None:
            raise RuntimeError(f'Stopwatch for {self._frame.name} is already running')
        self._start_time = time.monotonic()
        self._nvtx_range = nvtx.thread_range(self._frame._name_handle, domain=_domain)
        self._nvtx_range.__enter__()
        self._token = _current_frame.set(self._frame)

    def stop(self) -> None:
        if self._start_time is None:
            return
        assert self._token is not None
        _current_frame.reset(self._token)
        self._token = None
        self._nvtx_range.__exit__(None, None, None)
        stop_time = time.monotonic()
        record = Record(self._frame, self._start_time, stop_time)
        self._profiler.add_record(record)
        self._start_time = None


class NullStopwatch(Stopwatch):
    """Stopwatch that doesn't do anything."""

    def start(self) -> None:
        if self._start_time is not None:
            raise RuntimeError(f'Stopwatch {self._frame.name} is already running')
        self._start_time = 0.0   # Just for the error checking

    def stop(self) -> None:
        self._start_time = None


class Profiler(ABC):
    """Top-level class to hold profiling results."""

    @abstractmethod
    def add_record(self, record: Record) -> None:
        """Add a single event record to the results."""

    @abstractmethod
    def add_device_record(self, record: DeviceRecord) -> None:
        """Add a single device event record to the results."""

    @contextlib.contextmanager
    def profile(self, name: str,
                labels: Mapping[str, Any] = {}) -> Generator[Stopwatch, None, None]:
        """Context manager that runs code under a :class:`Stopwatch` with a new Frame."""
        frame = Frame(name, labels, _current_frame.get())
        with Stopwatch(self, frame) as stopwatch:
            yield stopwatch

    @contextlib.contextmanager
    def profile_device(self, queue: katsdpsigproc.abc.AbstractCommandQueue,
                       name: str, labels: Mapping[str, Any] = {}):
        frame = Frame(name, labels, _current_frame.get())
        start_event = queue.enqueue_marker()
        yield
        stop_event = queue.enqueue_marker()
        self.add_device_record(DeviceRecord(frame, start_event, stop_event))

    @staticmethod
    def get_profiler() -> 'Profiler':
        return _current_profiler.get()

    @staticmethod
    def set_profiler(profiler: 'Profiler') -> None:
        _current_profiler.set(profiler)


class CollectProfiler(Profiler):
    """Profiler that collects all the raw records.

    This can use an excessive amount of memory and is only intended for unit tests.
    """

    def __init__(self):
        self.records: List[Record] = []
        self.device_records: List[DeviceRecord] = []

    def add_record(self, record: Record) -> None:
        self.records.append(record)

    def add_device_record(self, record: DeviceRecord) -> None:
        self.device_records.append(record)


class FlamegraphProfiler(Profiler):
    """Profiler that summarises results by call stack to produce flame graphs.

    Specifically, use https://github.com/brendangregg/FlameGraph to
    post-process the output.

    Labels are discarded.
    """

    def __init__(self, device_record_buffer_size: int = 100000) -> None:
        self.flamegraph_samples: typing.Counter[Tuple[str, ...]] = collections.Counter()
        self.device_flamegraph_samples: typing.Counter[Tuple[str, ...]] = collections.Counter()
        # Querying device records immediately causes unwanted synchronisation
        # with the device. Keep a buffer to let them age.
        self.device_records: Deque[DeviceRecord] = deque()
        self.device_record_buffer_size = device_record_buffer_size

    def add_record(self, record: Record) -> None:
        stack = tuple(frame.name for frame in record.frame.stack())
        # flamegraph.pl wants integers, so convert to microseconds
        elapsed = round(1000000 * record.elapsed)
        self.flamegraph_samples[stack] += elapsed
        # We need to produce exclusive counts (time with no child frames
        # active), so subtract from parent frame.
        if len(stack) >= 2:
            self.flamegraph_samples[stack[:-1]] -= elapsed

    def _add_device_record(self, record: DeviceRecord) -> None:
        stack = tuple(frame.name for frame in record.frame.stack())
        # flamegraph.pl wants integers, so convert to microseconds
        elapsed = round(1000000 * record.elapsed)
        self.device_flamegraph_samples[stack] += elapsed

    def add_device_record(self, record: DeviceRecord) -> None:
        self.device_records.append(record)
        while len(self.device_records) > self.device_record_buffer_size:
            self._add_device_record(self.device_records.popleft())

    def write_flamegraph(self, file: Union[TextIO, io.TextIOBase]) -> None:
        """Writes host data that can be read by flamegraph.pl."""
        for names, value in self.flamegraph_samples.items():
            print(';'.join(names), value, file=file)

    def write_device_flamegraph(self, file: Union[TextIO, io.TextIOBase]) -> None:
        """Writes device data that can be read by flamegraph.pl."""
        while self.device_records:
            self._add_device_record(self.device_records.popleft())
        for names, value in self.device_flamegraph_samples.items():
            print(';'.join(names), value, file=file)


class NullProfiler(Profiler):
    """Implements the :class:`Profiler` interface but does not record anything."""

    def add_record(self, record: Record) -> None:
        pass

    def add_device_record(self, record: DeviceRecord) -> None:
        pass

    @contextlib.contextmanager
    def profile(self, name: str,
                labels: Mapping[str, Any] = {}) -> Generator[Stopwatch, None, None]:
        yield NullStopwatch(self, Frame(name, {}))

    @contextlib.contextmanager
    def profile_device(self, queue: katsdpsigproc.abc.AbstractCommandQueue,
                       name: str, labels: Mapping[str, Any] = {}):
        yield


# Create a default profiler that any context should inherit
_current_profiler: ContextVar[Profiler] = ContextVar('_current_profiler', default=NullProfiler())
_current_frame: ContextVar[Optional[Frame]] = ContextVar('_current_frame', default=None)


@contextlib.contextmanager
def profile(name: str, labels: Mapping[str, Any] = {}) -> Generator[Stopwatch, None, None]:
    """Context manager that runs code under a :class:`Stopwatch`.

    See also
    --------
    :meth:`Profiler.profile`
    """
    with Profiler.get_profiler().profile(name, labels) as stopwatch:
        yield stopwatch


def _extract_labels(sig: inspect.Signature, labels: _LabelSpec, *args, **kwargs):
    bound = sig.bind(*args, **kwargs)
    bound.apply_defaults()
    result = {}
    if isinstance(labels, collections.abc.Mapping):
        for label, spec in labels.items():
            if callable(spec):
                result[label] = spec(bound)
            else:
                result[label] = bound.arguments[spec]
    else:
        result = {label: bound.arguments[label] for label in labels}
    return result


def profile_function(name: Optional[str] = None, labels: _LabelSpec = ()) \
        -> Callable[[Callable], Callable]:
    """Decorator for profiling functions.

    It defaults to using the name of the function as the name for the frame,
    and the labels are extracted from the arguments passed to the function.
    The labels can either be a sequence or a mapping. If a sequence, each
    element names a function argument to be used as a label. If a mapping,
    the keys define the labels. Each values can either be the name of a
    function argument, or a callable that takes
    :class:`inspect.BoundArguments` and returns the value for the label.
    """
    def decorator(func: Callable) -> Callable:
        sig = inspect.signature(func)
        try:
            func_name = f'{func.__module__}.{func.__qualname__}'
        except AttributeError:
            func_name = '<anonymous function>'
        frame_name = name if name is not None else func_name

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            label_dict = _extract_labels(sig, labels, *args, **kwargs)
            with profile(frame_name, label_dict):
                return func(*args, **kwargs)

        return wrapper
    return decorator


def profile_generator(name: Optional[str] = None, labels: _LabelSpec = ()) \
        -> Callable[[Callable[..., Generator[_T, None, None]]],
                    Callable[..., Generator[_T, None, None]]]:
    """Decorator for profiling generator functions.

    It functions similarly to :func:`profile_function`.

    It is only suitable for simple generators intended to be used as
    iterables. It does not support sending values or exceptions into the
    wrapped generator.

    The stopwatch is stopped when the generator yields and restarted when it
    resumes.
    """

    labels = list(labels)

    def decorator(func: Callable) -> Callable:
        sig = inspect.signature(func)
        try:
            func_name = f'{func.__module__}.{func.__qualname__}'
        except AttributeError:
            func_name = '<anonymous function>'
        frame_name = name if name is not None else func_name

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            label_dict = _extract_labels(sig, labels, *args, **kwargs)
            gen = func(*args, **kwargs)
            try:
                with profile(frame_name, label_dict) as stopwatch:
                    for item in gen:
                        stopwatch.stop()
                        yield item
                        stopwatch.start()
            finally:
                gen.close()

        return wrapper
    return decorator


@contextlib.contextmanager
def profile_device(queue: katsdpsigproc.abc.AbstractCommandQueue,
                   name: str, labels: Mapping[str, Any] = {}):
    with Profiler.get_profiler().profile_device(queue, name, labels):
        yield



"""Build the _nvtx module."""

from cffi import FFI


ffibuilder = FFI()
ffibuilder.set_unicode(True)
ffibuilder.set_source(
    "._nvtx", "#include <nvtx3/nvToolsExt.h>\n",
    include_dirs=['/usr/local/cuda/include']
)
ffibuilder.cdef(
    r"""
    typedef ...* nvtxDomainHandle_t;
    typedef ...* nvtxStringHandle_t;
    #define NVTX_VERSION ...
    #define NVTX_EVENT_ATTRIB_STRUCT_SIZE ...

    typedef union {
        const char *ascii;
        const wchar_t *unicode;
        nvtxStringHandle_t registered;
        ...;
    } nvtxMessageValue_t;

    typedef enum {
        NVTX_COLOR_UNKNOWN,
        NVTX_COLOR_ARGB,
        ...
    } nvtxColorType_t;

    typedef enum {
        NVTX_MESSAGE_UNKNOWN,
        NVTX_MESSAGE_TYPE_ASCII,
        NVTX_MESSAGE_TYPE_UNICODE,
        NVTX_MESSAGE_TYPE_REGISTERED,
        ...
    } nvtxMessageType_t;

    typedef enum {
        NVTX_PAYLOAD_UNKNOWN,
        NVTX_PAYLOAD_TYPE_UNSIGNED_INT64,
        NVTX_PAYLOAD_TYPE_INT64,
        NVTX_PAYLOAD_TYPE_DOUBLE,
        NVTX_PAYLOAD_TYPE_UNSIGNED_INT32,
        NVTX_PAYLOAD_TYPE_INT32,
        NVTX_PAYLOAD_TYPE_FLOAT,
        ...
    } nvtxPayloadType_t;

    typedef struct {
        uint16_t version;
        uint16_t size;
        uint32_t category;
        int32_t colorType;
        uint32_t color;
        int32_t payloadType;
        union payload_t
        {
            uint64_t ullValue;
            int64_t llValue;
            double dValue;
            uint32_t uiValue;
            int32_t iValue;
            float fValue;
        } payload;
        int32_t messageType;
        nvtxMessageValue_t message;
        ...;
    } nvtxEventAttributes_t;

    void nvtxInitialize(const void *reserved);
    nvtxStringHandle_t nvtxDomainRegisterStringW(nvtxDomainHandle_t domain, const wchar_t* string);
    int nvtxDomainRangePushEx(nvtxDomainHandle_t domain, const nvtxEventAttributes_t *eventAttrib);
    int nvtxRangePushEx(const nvtxEventAttributes_t *eventAttrib);
    int nvtxRangePushW(const wchar_t *message);
    int nvtxDomainRangePop(nvtxDomainHandle_t domain);
    int nvtxRangePop(void);
    nvtxDomainHandle_t nvtxDomainCreateW(const wchar_t *name);
    void nvtxDomainDestroy(nvtxDomainHandle_t domain);
    """
)

if __name__ == '__main__':
    ffibuilder.compile(verbose=True)



"""Tests for :mod:`katsdpimager.arguments`"""

import argparse

import astropy.units as u
from nose.tools import assert_equal

from .. import arguments


class TestUnparse:
    def setup(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('--int', type=int)
        parser.add_argument('--int-default', type=int, default=3)
        parser.add_argument('--float', type=float)
        parser.add_argument('--float-default', type=float, default=2.5)
        parser.add_argument('--str', type=str)
        parser.add_argument('--str-default', type=str, default='')
        parser.add_argument('--positive', action='store_true')
        parser.add_argument('--no-negative', dest='negative', action='store_false')
        parser.add_argument('--quantity', type=u.Quantity)
        self.parser = parser

    def test_no_args(self):
        """Pass no arguments."""
        args = self.parser.parse_args([], namespace=arguments.SmartNamespace())
        assert_equal(arguments.unparse_args(args), [])

    def test_defaults(self):
        """Pass arguments with default values."""
        argv = ['--int-default=3', '--float-default=2.5', '--str-default=']
        args = self.parser.parse_args(argv, namespace=arguments.SmartNamespace())
        assert_equal(arguments.unparse_args(args), [])

    def test_all_args(self):
        """Pass all arguments."""
        argv = [
            '--int=1',
            '--int-default=2',
            '--float=3.1',
            '--float-default=4.5',
            '--str=hello',
            '--str-default=world',
            '--positive',
            '--no-negative',
            '--quantity=12.0 rad'
        ]
        args = self.parser.parse_args(argv, namespace=arguments.SmartNamespace())
        assert_equal(arguments.unparse_args(args), argv)

    def test_exclude(self):
        """Test the exclude argument."""
        self.parser.add_argument('positional', type=str)
        argv = ['positional', '--int=1', '--int-default=2']
        args = self.parser.parse_args(argv, namespace=arguments.SmartNamespace())
        unparsed = arguments.unparse_args(args, exclude=['positional', 'int'])
        assert_equal(unparsed, ['--int-default=2'])

    def test_arg_handler(self):
        """Test the `arg_handlers` option."""
        self.parser.add_argument('positional', type=str)
        argv = ['--int=1', 'positional_arg']
        args = self.parser.parse_args(argv, namespace=arguments.SmartNamespace())
        unparsed = arguments.unparse_args(
            args, arg_handlers={'positional': lambda name, value: [value]})
        assert_equal(unparsed, argv)

    def test_type_handler(self):
        """Test the type_handlers option."""
        self.parser.add_argument('--list', type=str, action='append')
        argv = ['--list=1', '--list=2']
        args = self.parser.parse_args(argv, namespace=arguments.SmartNamespace())
        unparsed = arguments.unparse_args(
            args, type_handlers={list: lambda name, value: [f'--{name}={x}' for x in value]})
        assert_equal(unparsed, argv)



"""Fake version of :mod:`katsdpimager._nvtx` that does nothing."""

import contextlib
from typing import Optional, Generator, Union


def initialize() -> None:
    pass


class RegisteredString:
    def __init__(self, handle: object) -> None:
        pass


class Domain:
    def __init__(self, name: str) -> None:
        pass


def register_string(value: str, domain: Optional[Domain] = None) -> RegisteredString:
    return RegisteredString(object())


@contextlib.contextmanager
def thread_range(name: Union[str, RegisteredString], *,
                 payload: Union[None, int, float] = None,
                 domain: Optional[Domain] = None) -> Generator[None, None, None]:
    yield



#!/usr/bin/env python3
import sys
import argparse
import logging
import shlex
from contextlib import closing, contextmanager

import numpy as np
import colors
import katsdpsigproc.accel as accel
import astropy.io.fits as fits

from katsdpimager import frontend, loader, io, progress, arguments, profiling


logger = logging.getLogger()


class Writer(frontend.Writer):
    def __init__(self, args, dataset):
        self.args = args
        options = frontend.command_line_options(
            args, dataset, {'input_file', 'output_file', 'input_option'})
        self.extra_fits_headers = fits.Header()
        self.extra_fits_headers['HISTORY'] = \
            'Command line options: ' + ' '.join(shlex.quote(arg) for arg in options)

    def needs_fits_image(self, name):
        if name == 'clean':
            return True
        else:
            return getattr(self.args, 'write_' + name) is not None

    def needs_fits_grid(self, name):
        return getattr(self.args, 'write_' + name) is not None

    def write_fits_image(self, name, description, dataset, image, image_parameters, channel,
                         beam=None, bunit='Jy/beam'):
        if name == 'clean':
            filename = self.args.output_file
        else:
            filename = getattr(self.args, 'write_' + name)
        if filename is not None:
            if '%' in filename:
                filename = filename % channel
            with progress.step('Write {}'.format(description)):
                io.write_fits_image(dataset, image, image_parameters, filename,
                                    channel, beam, bunit, self.extra_fits_headers)

    def write_fits_grid(self, name, description, fftshift, grid_data, image_parameters, channel):
        filename = getattr(self.args, 'write_' + name)
        if filename is not None:
            if '%' in filename:
                filename = filename % channel
            with progress.step('Write {}'.format(description)):
                if fftshift:
                    grid_data = np.fft.fftshift(grid_data, axes=(1, 2))
                io.write_fits_grid(grid_data, image_parameters, filename, channel)


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('input_file', type=str, metavar='INPUT',
                        help='Input data set')
    parser.add_argument('output_file', type=str, metavar='OUTPUT',
                        help='Output FITS file')
    parser.add_argument('--log-level', type=str, default='INFO', metavar='LEVEL',
                        help='Logging level [%(default)s]')

    frontend.add_options(parser)

    group = parser.add_argument_group('Debugging options')
    group.add_argument('--host', action='store_true',
                       help='Perform operations on the CPU')
    group.add_argument('--write-weights', metavar='FILE',
                       help='Write imaging weights to FITS file')
    group.add_argument('--write-psf', metavar='FILE',
                       help='Write image of PSF to FITS file')
    group.add_argument('--write-grid', metavar='FILE',
                       help='Write UV grid to FITS file')
    group.add_argument('--write-dirty', metavar='FILE',
                       help='Write dirty image to FITS file')
    group.add_argument('--write-model', metavar='FILE',
                       help='Write model image to FITS file')
    group.add_argument('--write-residuals', metavar='FILE',
                       help='Write image residuals to FITS file')
    group.add_argument('--write-primary-beam', metavar='FILE',
                       help='Write primary beam model to FITS file')
    group.add_argument('--write-profile', metavar='FILE',
                       help='Write profiling information to file')
    group.add_argument('--write-device-profile', metavar='FILE',
                       help='Write device profiling information to file')
    group.add_argument('--vis-limit', type=int, metavar='N',
                       help='Use only the first N visibilities')
    return parser


class DummyCommandQueue:
    """Stub equivalent to CUDA/OpenCL CommandQueue objects, that just provides
    an empty :meth:`finish` method."""
    def finish(self):
        pass


@contextmanager
def dummy_context():
    """Do-nothing context manager used in place of a device context."""
    yield


class ColorFormatter(logging.Formatter):
    COLORS = {
        logging.CRITICAL: colors.red,
        logging.ERROR: colors.red,
        logging.WARNING: colors.magenta,
        logging.INFO: colors.green,
        logging.DEBUG: colors.cyan
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def format(self, record):
        msg = super().format(record)
        if record.levelno in self.COLORS:
            msg = self.COLORS[record.levelno](msg)
        return msg


def configure_logging(args):
    log_handler = logging.StreamHandler()
    fmt = "[%(levelname)s] %(message)s"
    if sys.stderr.isatty():
        log_handler.setFormatter(ColorFormatter(fmt))
    else:
        log_handler.setFormatter(logging.Formatter(fmt))
    # In case some other module has decided to configure logging, remove any
    # existing handlers.
    for handler in list(logger.handlers):
        logger.removeHandler(handler)
    logger.addHandler(log_handler)
    logger.setLevel(args.log_level.upper())


def main():
    parser = get_parser()
    args = parser.parse_args(namespace=arguments.SmartNamespace())
    if (args.stop_channel is None or args.stop_channel - args.start_channel > 1):
        if '%' not in args.output_file:
            parser.error('More than one channel selected but no %d in output filename')
    configure_logging(args)
    if args.write_profile or args.write_device_profile:
        profiling.Profiler.set_profiler(profiling.FlamegraphProfiler())

    queue = None
    context = None
    if not args.host:
        context = accel.create_some_context(device_filter=lambda x: x.is_cuda)
        queue = context.create_command_queue()
    else:
        context = dummy_context()
        queue = DummyCommandQueue()

    with closing(loader.load(args.input_file, args.input_option,
                             args.start_channel, args.stop_channel)) as dataset:
        frontend.run(args, context, queue, dataset, Writer(args, dataset))

    profiler = profiling.Profiler.get_profiler()
    if args.write_profile:
        with open(args.write_profile, 'w') as f:
            assert isinstance(profiler, profiling.FlamegraphProfiler)
            profiler.write_flamegraph(f)
    if args.write_device_profile:
        with open(args.write_device_profile, 'w') as f:
            assert isinstance(profiler, profiling.FlamegraphProfiler)
            profiler.write_device_flamegraph(f)


if __name__ == '__main__':
    main()



"""Top-level objects for accelerated imaging, that glue the lower-level
objects together."""

import numpy as np
from katsdpsigproc import accel

from . import grid, predict, weight, image, clean
from .profiling import profile_function, profile_device


class ImagingTemplate:
    """Template holding all the other templates for imaging."""

    @profile_function()
    def __init__(self, context, array_parameters, fixed_image_parameters,
                 weight_parameters, fixed_grid_parameters, clean_parameters):
        self.context = context
        self.array_parameters = array_parameters
        self.fixed_image_parameters = fixed_image_parameters
        self.weight_parameters = weight_parameters
        self.fixed_grid_parameters = fixed_grid_parameters
        self.clean_parameters = clean_parameters
        num_polarizations = len(fixed_image_parameters.polarizations)
        self.weights = weight.WeightsTemplate(
            context, weight_parameters.weight_type, num_polarizations)
        self.gridder = grid.GridderTemplate(
            context, fixed_image_parameters, fixed_grid_parameters)
        self.predict = predict.PredictTemplate(
            context, fixed_image_parameters.real_dtype, num_polarizations)
        self.grid_image = image.GridImageTemplate(
            context, fixed_image_parameters.real_dtype)
        self.psf_patch = clean.PsfPatchTemplate(
            context, fixed_image_parameters.real_dtype, num_polarizations)
        self.noise_est = clean.NoiseEstTemplate(
            context, fixed_image_parameters.real_dtype, num_polarizations)
        self.clean = clean.CleanTemplate(
            context, clean_parameters, fixed_image_parameters.real_dtype, num_polarizations)
        self.scale = image.ScaleTemplate(
            context, fixed_image_parameters.real_dtype, num_polarizations)
        self.add_image = image.AddImageTemplate(
            context, fixed_image_parameters.real_dtype, num_polarizations)
        self.apply_primary_beam = image.ApplyPrimaryBeamTemplate(
            context, fixed_image_parameters.real_dtype, num_polarizations)
        if fixed_grid_parameters.degrid:
            self.degridder = grid.DegridderTemplate(
                context, fixed_image_parameters, fixed_grid_parameters)
        else:
            self.degridder = None

    def instantiate(self, *args, **kwargs):
        return Imaging(self, *args, **kwargs)


class _HostBuffer:
    """A host array with supporting utilities for synchronisation."""

    def __init__(self, context, slot):
        self.array = accel.HostArray(slot.shape, slot.dtype, slot.required_padded_shape(),
                                     context=context)
        self.transfer_event = None


def _get_uv(coords):
    """Extract view of uv and sub_uv fields as an Nx4 array."""
    # Check that types are as expected and contiguous
    assert coords.dtype['uv'] == np.dtype(('i2', (2,)))
    assert coords.dtype['sub_uv'] == np.dtype(('i2', (2,)))
    assert (coords.dtype.fields['sub_uv'][1]
            == coords.dtype.fields['uv'][1] + coords.dtype['uv'].itemsize)
    # Create a dtype alias where these are just a single field
    new_dtype = np.dtype(dict(
        names=['uv_sub_uv'],
        formats=[('i2', (4,))],
        offsets=[coords.dtype.fields['uv'][1]],
        itemsize=coords.dtype.itemsize
    ))
    alias = coords.view(new_dtype)
    return alias['uv_sub_uv']


class Imaging(accel.OperationSequence):
    @profile_function()
    def __init__(
            self, template, command_queue,
            image_parameters, grid_parameters,
            max_vis, max_sources, major, allocator=None):
        assert image_parameters.fixed == template.fixed_image_parameters
        assert grid_parameters.fixed == template.fixed_grid_parameters
        self.template = template
        lm_scale = float(image_parameters.pixel_size)
        lm_bias = -0.5 * image_parameters.pixels * lm_scale
        image_shape = (len(image_parameters.fixed.polarizations),
                       image_parameters.pixels,
                       image_parameters.pixels)
        # Currently none of the kernels accessing the layer need any padding.
        # It would be nice if there was a cleaner way to handle this; possibly
        # by deferring creation of the FFT plan until instantiation.
        padded_layer_shape = layer_shape = image_shape[1:]
        fft_plan = template.grid_image.make_fft_plan(layer_shape, padded_layer_shape)

        self._gridder = template.gridder.instantiate(
            command_queue,
            template.array_parameters,
            image_parameters,
            grid_parameters,
            max_vis,
            allocator)
        self._continuum_predict = template.predict.instantiate(
            command_queue, image_parameters, grid_parameters,
            max_vis, max_sources, allocator)
        grid_shape = self._gridder.slots['grid'].shape
        self._weights = template.weights.instantiate(
            command_queue, grid_shape, max_vis, allocator)
        self._weights.robustness = template.weight_parameters.robustness
        self._grid_to_image = template.grid_image.instantiate_grid_to_image(
            command_queue, grid_shape, lm_scale, lm_bias, fft_plan, allocator)
        self._psf_patch = template.psf_patch.instantiate(
            command_queue, image_shape, allocator)
        self._noise_est = template.noise_est.instantiate(
            command_queue, image_shape, template.clean_parameters.border, allocator)
        self._clean = template.clean.instantiate(
            command_queue, image_parameters, allocator)
        self._scale = template.scale.instantiate(
            command_queue, image_shape, allocator)
        self._add_image = template.add_image.instantiate(
            command_queue, image_shape, allocator)
        # Thresholds are set later by self.apply_primary_beam
        self._apply_primary_beam_model = template.apply_primary_beam.instantiate(
            command_queue, image_shape, 0.0, 0.0, allocator)
        self._apply_primary_beam_dirty = template.apply_primary_beam.instantiate(
            command_queue, image_shape, 0.0, np.nan, allocator)

        # TODO: handle taper1d/untaper1d as standard slots
        taper1d = accel.DeviceArray(
            template.context,
            (image_parameters.pixels,),
            image_parameters.fixed.real_dtype)
        taper1d_host = taper1d.empty_like()
        self._gridder.convolve_kernel.taper(image_parameters.pixels, taper1d_host)
        taper1d.set(command_queue, taper1d_host)
        self._grid_to_image.bind(kernel1d=taper1d)
        del taper1d_host
        if grid_parameters.fixed.degrid:
            untaper1d = accel.DeviceArray(
                template.context,
                (image_parameters.pixels,),
                image_parameters.fixed.real_dtype)
            untaper1d_host = untaper1d.empty_like()
            self._predict = template.degridder.instantiate(
                command_queue,
                template.array_parameters,
                image_parameters,
                grid_parameters,
                max_vis,
                allocator)
            self._predict.convolve_kernel.taper(image_parameters.pixels, untaper1d_host)
            untaper1d.set(command_queue, untaper1d_host)
            del untaper1d_host

            degrid_shape = self._predict.slots['grid'].shape
            self._image_to_grid = template.grid_image.instantiate_image_to_grid(
                command_queue, degrid_shape, lm_scale, lm_bias, fft_plan, allocator)
            self._image_to_grid.bind(kernel1d=untaper1d)
        else:
            max_components = min(image_parameters.pixels**2,
                                 (major - 1) * template.clean_parameters.minor)
            self._predict = template.predict.instantiate(
                command_queue, image_parameters, grid_parameters,
                max_vis, max_components, allocator)
        self._model_components = {}
        operations = [
            ('weights', self._weights),
            ('gridder', self._gridder),
            ('predict', self._predict),
            ('continuum_predict', self._continuum_predict),
            ('grid_to_image', self._grid_to_image),
            ('psf_patch', self._psf_patch),
            ('noise_est', self._noise_est),
            ('clean', self._clean),
            ('scale', self._scale),
            ('add_image', self._add_image),
            ('apply_primary_beam_model', self._apply_primary_beam_model),
            ('apply_primary_beam_dirty', self._apply_primary_beam_dirty)
        ]
        compounds = {
            'weights': ['weights:weights', 'predict:weights', 'continuum_predict:weights'],
            'weights_grid': ['weights:grid', 'gridder:weights_grid'],
            'uv': ['weights:uv', 'gridder:uv', 'predict:uv', 'continuum_predict:uv'],
            'w_plane': ['gridder:w_plane', 'predict:w_plane', 'continuum_predict:w_plane'],
            'vis': ['gridder:vis', 'predict:vis', 'continuum_predict:vis'],
            'grid': ['gridder:grid', 'grid_to_image:grid'],
            'layer': ['grid_to_image:layer'],
            'dirty': ['grid_to_image:image', 'noise_est:dirty', 'clean:dirty', 'scale:data',
                      'add_image:dest', 'apply_primary_beam_dirty:data'],
            'model': ['clean:model', 'apply_primary_beam_model:data', 'add_image:src'],
            'psf': ['clean:psf', 'psf_patch:psf'],
            'tile_max': ['clean:tile_max'],
            'tile_pos': ['clean:tile_pos'],
            'peak_value': ['clean:peak_value'],
            'peak_pos': ['clean:peak_pos'],
            'peak_pixel': ['clean:peak_pixel'],
            'beam_power': ['apply_primary_beam_model:beam_power',
                           'apply_primary_beam_dirty:beam_power']
        }
        if grid_parameters.fixed.degrid:
            operations.append(('image_to_grid', self._image_to_grid))
            compounds['degrid'] = ['predict:grid', 'image_to_grid:grid']
            compounds['layer'].append('image_to_grid:layer')
            compounds['model'].append('image_to_grid:image')
        # TODO: could alias weights with something, since it's only needed
        # early on while setting up weights_grid.
        # TODO: could alias noise_est:rank with something, since it's only
        # needed at the start of the minor cycles (it is small though).
        super().__init__(
            command_queue, operations, compounds, allocator=allocator)
        # dirty_to_psf swaps the dirty and PSF images, so they need to have
        # compatible padding.
        for x, y in zip(self.slots['dirty'].dimensions, self.slots['psf'].dimensions):
            x.link(y)

        self.host_buffer = {}
        for name in ['weights', 'uv', 'w_plane', 'vis']:
            if name in self.slots:
                self.host_buffer[name] = _HostBuffer(command_queue.context, self.slots[name])

    def __call__(self, **kwargs):
        raise NotImplementedError()

    @property
    def num_vis(self):
        return self._gridder.num_vis

    @num_vis.setter
    def num_vis(self, value):
        self._gridder.num_vis = value
        self._predict.num_vis = value
        self._continuum_predict.num_vis = value

    @profile_function()
    def clear_weights(self):
        self._weights.clear()

    @profile_function()
    def grid_weights(self, uv, weights):
        self._set_buffer('uv', len(uv), uv, (np.s_[:2],))
        self._set_buffer('weights', len(uv), weights)
        self._weights.grid(len(uv))

    @profile_function()
    def finalize_weights(self):
        return self._weights.finalize()

    @profile_function()
    def clear_grid(self):
        with profile_device(self.command_queue, 'clear_grid'):
            self.buffer('grid').zero(self.command_queue)

    @profile_function()
    def clear_dirty(self):
        with profile_device(self.command_queue, 'clear_dirty'):
            self.buffer('dirty').zero(self.command_queue)

    @profile_function()
    def clear_model(self):
        with profile_device(self.command_queue, 'clear_model'):
            self.buffer('model').zero(self.command_queue)
        self._model_components.clear()

    def _set_buffer(self, name, N, data, extra_index=()):
        if len(data) != N:
            raise ValueError('Lengths do not match')
        host = self.host_buffer[name]
        device = self.buffer(name)
        if host.transfer_event is not None:
            host.transfer_event.wait()
        idx = (np.s_[:N],) + extra_index
        host.array[idx] = data
        device.set_region(self.command_queue, host.array, idx, idx, blocking=False)
        host.transfer_event = self.command_queue.enqueue_marker()

    def _set_uv(self, coords):
        N = self.num_vis
        if len(coords) != N:
            raise ValueError('Lengths do not match')
        host = self.host_buffer['uv']
        device = self.buffer('uv')
        if host.transfer_event is not None:
            host.transfer_event.wait()
        host.array[:N] = _get_uv(coords)
        device.set_region(self.command_queue, host.array, np.s_[:N], np.s_[:N], blocking=False)
        host.transfer_event = self.command_queue.enqueue_marker()

    @profile_function()
    def set_coordinates(self, coords):
        """Set UVW coordinates.

        Parameters
        ----------
        coords
            Structured array with fields ``uv``, ``sub_uv`` and ``w_plane``. The
            former two must each be a 2-element int16 array, and they must be
            contiguous in the structure.
        """
        self._set_uv(coords)
        self._set_buffer('w_plane', self.num_vis, coords['w_plane'])

    @profile_function()
    def set_vis(self, vis):
        self._set_buffer('vis', self.num_vis, vis)

    @profile_function()
    def set_weights(self, weights):
        """Set statistical weights for prediction"""
        self._set_buffer('weights', self.num_vis, weights)

    @profile_function()
    def grid(self):
        self._gridder()

    @profile_function()
    def predict(self, w):
        if not self.template.fixed_grid_parameters.degrid:
            self._predict.set_w(w)
        self._predict()

    @profile_function()
    def continuum_predict(self, w):
        self._continuum_predict.set_w(w)
        self._continuum_predict()

    def set_sky_model(self, sky_model, phase_centre):
        self._continuum_predict.set_sky_model(sky_model, phase_centre)

    @profile_function()
    def grid_to_image(self, w):
        self._grid_to_image.set_w(w)
        self._grid_to_image()

    @profile_function()
    def model_to_grid(self, w):
        if not self._image_to_grid:
            raise RuntimeError('Can only use model_to_grid with degridding')
        self._image_to_grid.set_w(w)
        self._image_to_grid()

    @profile_function()
    def model_to_predict(self):
        if self.template.fixed_grid_parameters.degrid:
            raise RuntimeError('Can only use model_to_predict with direct prediction')
        self._predict.set_sky_image(self._model_components)

    @profile_function()
    def scale_dirty(self, scale_factor):
        self._scale.set_scale_factor(scale_factor)
        self._scale()

    @profile_function()
    def add_model_to_dirty(self):
        self._add_image()

    @profile_function()
    def apply_primary_beam(self, threshold):
        """Applies primary beam power to both model and dirty images."""
        self._apply_primary_beam_model.threshold = threshold
        self._apply_primary_beam_model()
        self._apply_primary_beam_dirty.threshold = threshold
        self._apply_primary_beam_dirty()

    @profile_function()
    def dirty_to_psf(self):
        dirty = self.buffer('dirty')
        psf = self.buffer('psf')
        self.bind(dirty=psf, psf=dirty)

    @profile_function()
    def psf_patch(self):
        return self._psf_patch(self.template.clean_parameters.psf_cutoff,
                               self.template.clean_parameters.psf_limit)

    @profile_function()
    def noise_est(self):
        return self._noise_est()

    @profile_function()
    def clean_reset(self):
        self._clean.reset()

    @profile_function()
    def clean_cycle(self, psf_patch, threshold=0.0):
        peak_value, peak_pos, model_pixel = self._clean(psf_patch, threshold)
        if peak_pos is not None:
            try:
                self._model_components[peak_pos] += model_pixel
            except KeyError:
                self._model_components[peak_pos] = model_pixel
        return peak_value

    @profile_function(labels=['name'])
    def get_buffer(self, name):
        """Get the contents of a buffer as a numpy array."""
        return self.buffer(name).get(self.command_queue)

    @profile_function(labels=['name'])
    def set_buffer(self, name, data):
        """Copy a numpy array to a buffer (blocking).

        Because this is blocking, it is not considered a high performance
        path. It exists to simplify compatibility with :class:`ImagingHost`.
        """
        self.buffer(name).set(self.command_queue, data)

    def free_buffer(self, name):
        """Free the memory associated with a given buffer.

        It is the caller's responsibility to ensure that the buffer is not
        needed in future operations.
        """
        if name in self.slots:
            self.slots[name].bind(None)


class ImagingHost:
    """Host-only equivalent to :class:`Imaging`."""

    def __init__(self, image_parameters, weight_parameters, grid_parameters, clean_parameters):
        lm_scale = float(image_parameters.pixel_size)
        lm_bias = -0.5 * image_parameters.pixels * lm_scale
        self._clean_parameters = clean_parameters
        self._image_parameters = image_parameters
        self._gridder = grid.GridderHost(image_parameters, grid_parameters)
        self._grid = self._gridder.values
        self._continuum_predict = predict.PredictHost(image_parameters, grid_parameters)
        self._weights_grid = self._gridder.weights_grid
        self._weights = weight.WeightsHost(weight_parameters.weight_type, self._weights_grid)
        self._weights.robustness = weight_parameters.robustness
        self._layer = np.empty(self._grid.shape, image_parameters.fixed.complex_dtype)
        self._dirty = np.empty(self._grid.shape, image_parameters.fixed.real_dtype)
        self._model = np.empty(self._grid.shape, image_parameters.fixed.real_dtype)
        self._psf = np.empty(self._grid.shape, image_parameters.fixed.real_dtype)
        self._beam_power = np.empty(self._grid.shape[1:], image_parameters.fixed.real_dtype)
        self._grid_to_image = image.GridToImageHost(
            self._grid, self._layer, self._dirty,
            self._gridder.kernel.taper(image_parameters.pixels), lm_scale, lm_bias)
        self._clean = clean.CleanHost(image_parameters, clean_parameters,
                                      self._dirty, self._psf, self._model)
        self._model_components = {}
        if grid_parameters.fixed.degrid:
            self._predict = grid.DegridderHost(image_parameters, grid_parameters)
            self._degrid = self._predict.values
            self._image_to_grid = image.ImageToGridHost(
                self._degrid, self._layer, self._model,
                self._predict.kernel.taper(image_parameters.pixels), lm_scale, lm_bias)
        else:
            self._predict = predict.PredictHost(image_parameters, grid_parameters)
            self._degrid = None
            self._image_to_grid = None
        self._buffer = {
            'psf': self._psf,
            'dirty': self._dirty,
            'model': self._model,
            'grid': self._grid,
            'weights_grid': self._weights_grid,
            'beam_power': self._beam_power
        }
        if self._degrid is not None:
            self._buffer['degrid'] = self._degrid

    def buffer(self, name):
        return self._buffer[name]

    def get_buffer(self, name):
        return self._buffer[name]

    def set_buffer(self, name, data):
        self._buffer[name][()] = data

    def free_buffer(self, name):
        if name in self._buffer:
            self._buffer[name] = None

    @property
    def num_vis(self):
        return self._gridder.num_vis

    @num_vis.setter
    def num_vis(self, value):
        self._gridder.num_vis = value
        self._predict.num_vis = value
        self._continuum_predict.num_vis = value

    def clear_weights(self):
        self._weights_grid.fill(0)

    def grid_weights(self, uv, weights):
        self._weights.grid(uv, weights)

    def finalize_weights(self):
        return self._weights.finalize()

    def clear_grid(self):
        self._grid.fill(0)

    def clear_dirty(self):
        self._dirty.fill(0)

    def clear_model(self):
        self._model.fill(0)
        self._model_components.clear()

    def set_sky_model(self, sky_model, phase_centre):
        self._continuum_predict.set_sky_model(sky_model, phase_centre)

    def set_coordinates(self, coords):
        self._gridder.set_coordinates(coords.uv, coords.sub_uv, coords.w_plane)
        self._predict.set_coordinates(coords.uv, coords.sub_uv, coords.w_plane)
        self._continuum_predict.set_coordinates(coords.uv, coords.sub_uv, coords.w_plane)

    def set_vis(self, *args, **kwargs):
        self._gridder.set_vis(*args, **kwargs)
        self._predict.set_vis(*args, **kwargs)
        self._continuum_predict.set_vis(*args, **kwargs)

    def set_weights(self, *args, **kwargs):
        self._predict.set_weights(*args, **kwargs)
        self._continuum_predict.set_weights(*args, **kwargs)

    def grid(self):
        self._gridder()

    def predict(self, w):
        if self._degrid is None:
            self._predict.set_w(w)
        self._predict()

    def continuum_predict(self, w):
        self._continuum_predict.set_w(w)
        self._continuum_predict()

    def grid_to_image(self, w):
        self._grid_to_image.set_w(w)
        self._grid_to_image()

    def model_to_grid(self, w):
        if self._degrid is None:
            raise RuntimeError('Can only use model_to_grid with degridding')
        self._image_to_grid.set_w(w)
        self._image_to_grid()

    def model_to_predict(self):
        if self._degrid is not None:
            raise RuntimeError('Can only use model_to_predict with direct prediction')
        self._predict.set_sky_image(self._model_components)

    def scale_dirty(self, scale_factor):
        self._dirty *= scale_factor[:, np.newaxis, np.newaxis]

    def add_model_to_dirty(self):
        self._dirty += self._model

    def apply_primary_beam(self, threshold):
        mask = (self._beam_power < threshold)[np.newaxis, ...]
        self._model /= self._beam_power
        self._model[mask] = 0.0
        self._dirty /= self._beam_power
        self._dirty[mask] = np.nan

    def dirty_to_psf(self):
        self._psf[:] = self._dirty

    def psf_patch(self):
        return clean.psf_patch_host(self._psf,
                                    self._clean_parameters.psf_cutoff,
                                    self._clean_parameters.psf_limit)

    def noise_est(self):
        return clean.noise_est_host(self._dirty, self._clean_parameters.border)

    def clean_reset(self):
        self._clean.reset()

    def clean_cycle(self, psf_patch, threshold=0.0):
        peak_value, peak_pos, model_pixel = self._clean(psf_patch, threshold)
        if peak_pos is not None:
            try:
                self._model_components[peak_pos] += model_pixel
            except KeyError:
                self._model_components[peak_pos] = model_pixel
        return peak_value



#!/usr/bin/env python3
import argparse
import os
import logging
import shutil
import uuid

import katdal
import katsdpservices

from katsdpimager import report


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument('dataset', type=str, help='Input dataset')
    parser.add_argument('output_dir', type=str, help='Parent directory for output')
    parser.add_argument('prefix', type=str, help='Prefix for output directories and filenames')
    parser.add_argument('stream', type=str, help='Stream name for telescope state inputs')
    parser.add_argument('--log-level', type=str, metavar='LEVEL',
                        help='Logging level [INFO]')
    args = parser.parse_args()
    katsdpservices.setup_logging()
    if args.log_level is not None:
        logging.getLogger().setLevel(args.log_level.upper())

    dataset = katdal.open(args.dataset, chunk_store=None, upgrade_flags=False)
    telstate = dataset.source.telstate.wrapped.root()
    telstate = telstate.view(telstate.join(dataset.source.capture_block_id, args.stream))

    output_dir = '{}_{}'.format(args.prefix, uuid.uuid4())
    output_dir = os.path.join(args.output_dir, output_dir)
    filename = args.prefix + '_report.html'
    tmp_dir = output_dir + '.writing'
    os.mkdir(tmp_dir)
    try:
        common_stats, target_stats = report.get_stats(dataset, telstate)
        report.write_report(common_stats, target_stats, os.path.join(tmp_dir, filename))
        report.write_metadata(dataset, common_stats, target_stats,
                              os.path.join(tmp_dir, 'metadata.json'))
        os.rename(tmp_dir, output_dir)
    except Exception:
        # Make a best effort to clean up
        shutil.rmtree(tmp_dir, ignore_errors=True)
        raise


if __name__ == '__main__':
    main()



"""Tests for :py:mod:`katsdpimager.frontend`"""

import numpy as np

from .. import frontend


def test_find_peak():
    # Create something very vaguely like a primary beam.
    size = 4096
    noise = 15.0
    peak = 200.0

    x = np.linspace(-np.pi / 2, np.pi / 2, size)
    y = np.cos(x)
    pbeam = y[np.newaxis, :] * y[:, np.newaxis]
    pbeam[pbeam < 0.01] = np.nan

    rs = np.random.RandomState(seed=1)
    image = rs.normal(scale=noise, size=(4, size, size)) / pbeam

    np.testing.assert_equal(frontend.find_peak(image, pbeam, noise), np.nan)
    image[1, size // 2 + 5, size // 2 - 10] = peak
    np.testing.assert_equal(frontend.find_peak(image, pbeam, noise), peak)
    image *= -1    # Test negative peaks
    np.testing.assert_equal(frontend.find_peak(image, pbeam, noise), peak)



"""Tests for :mod:`katsdpimager.nvtx`.

It tests both the real and the fallback implementation. It is unable to test
that it actually inserts the intended values into the NVIDIA tools, but it
can check that nothing crashes and that the fallback implementation is
complete.
"""

from .. import nvtx, _nvtx_fallback


class TestNvtx:
    MOD = nvtx

    @classmethod
    def setupClass(cls) -> None:
        # Needs to be done only once for testing with nsys - creating and
        # destroying the same named domain causes problems otherwise.
        cls.domain = cls.MOD.Domain('test_domain')

    def setup(self) -> None:
        self.rstring1 = self.MOD.register_string('hello')
        self.rstring2 = self.MOD.register_string('world', domain=self.domain)

    def test_initialize(self) -> None:
        self.MOD.initialize()

    def test_thread_range_basic(self) -> None:
        with self.MOD.thread_range('basic'):
            pass

    def test_thread_range_payload(self) -> None:
        with self.MOD.thread_range('int', payload=1):
            with self.MOD.thread_range('float', payload=2.5):
                pass

    def test_thread_range_domain(self) -> None:
        with self.MOD.thread_range('domain', domain=self.domain):
            pass

    def test_thread_range_registered_string(self) -> None:
        with self.MOD.thread_range(self.rstring1):
            with self.MOD.thread_range(self.rstring2, domain=self.domain):
                pass


class TestNvtxFallback(TestNvtx):
    MOD = _nvtx_fallback



"""Create partial metadata.json for katsdptransfer"""

from datetime import datetime
from typing import Sequence, Dict

import numpy as np
import katpoint
import katdal
import astropy.units as u


def target_mask(dataset: katdal.DataSet, target: katpoint.Target) -> np.ndarray:
    """Get boolean mask of dumps where the target is being tracked."""
    # Based on katdal.DataSet.select, but we don't actually use select because
    # we don't want to modify the dataset.
    scan_sensor = dataset.sensor['Observation/scan_state']
    track_mask = (scan_sensor == 'track')
    target_index_sensor = dataset.sensor['Observation/target_index']
    target_index = dataset.catalogue.targets.index(target)
    mask = (target_index_sensor == target_index) & track_mask
    return mask


def time_on_target(dataset: katdal.DataSet, target: katpoint.Target) -> u.Quantity:
    """Seconds spent tracking the target.

    This might be a slight under-report, because dumps that span the
    boundaries are marked as not tracking by katdal, but may have partial
    data that was nevertheless usable.
    """
    mask = target_mask(dataset, target)
    return dataset.dump_period * np.sum(mask) * u.s


def make_metadata(dataset: katdal.DataSet, targets: Sequence[katpoint.Target],
                  channels: int,
                  description: str) -> Dict[str, object]:
    obs_params = dataset.obs_params
    return {
        'CaptureBlockId': dataset.source.capture_block_id,
        'ScheduleBlockIdCode': obs_params.get('sb_id_code', 'UNKNOWN'),
        'Description': obs_params.get('description', 'UNKNOWN') + ': ' + description,
        'ProposalId': obs_params.get('proposal_id', 'UNKNOWN'),
        'Observer': obs_params.get('observer', 'UNKNOWN'),
        # Solr doesn't accept +00:00, only Z, so we can't just format a timezone-aware value
        'StartTime': datetime.utcnow().isoformat() + 'Z',
        'Bandwidth': dataset.channel_width * channels,
        'ChannelWidth': dataset.channel_width,
        'NumFreqChannels': channels,
        'RightAscension': [str(target.radec()[0]) for target in targets],
        'Declination': [str(target.radec()[1]) for target in targets],
        # JSON schema limits format to fixed-point with at most 10 decimal places
        'DecRa': [
            ','.join('{:.10f}'.format(a) for a in np.rad2deg(target.radec())[::-1])
            for target in targets
        ],
        'Targets': [target.name for target in targets],
        'KatpointTargets': [target.description for target in targets],
        # metadata value needs to be in hours
        'IntegrationTime': [time_on_target(dataset, target).to_value(u.h)
                            for target in targets]
    }



"""Tests for :py:mod:`katsdpimager.report`"""

import astropy.units as u
import numpy as np

from .. import report


class TestPolynomialSEFDModel:
    def setup(self):
        self.model_simple = report.PolynomialSEFDModel(
            2 * u.MHz, 10 * u.MHz,
            [[3, 1, 2]] * u.Jy, u.MHz, 0.5)
        self.model_mixed_units = report.PolynomialSEFDModel(
            2000 * u.kHz, 1e7 * u.Hz,
            [[3000, 1000, 2000]] * u.mJy, u.MHz, 0.5)
        self.freq = [1, 2, 3, 10, 11] * u.MHz

    def test_simple(self):
        sefd = self.model_simple(self.freq).to_value(u.Jy)
        np.testing.assert_allclose(sefd, [np.nan, 13.0, 24.0, 213.0, np.nan])

    def test_mixed_units(self):
        sefd = self.model_mixed_units(self.freq.to(u.GHz)).to_value(u.Jy)
        np.testing.assert_allclose(sefd, [np.nan, 13.0, 24.0, 213.0, np.nan])

    def test_effective(self):
        sefd = self.model_simple(self.freq).to_value(u.Jy)
        sefd_effective = self.model_simple(self.freq, effective=True).to_value(u.Jy)
        np.testing.assert_allclose(sefd_effective, sefd * 2)

    def test_multi_pol(self):
        model = report.PolynomialSEFDModel(
            2 * u.MHz, 4 * u.MHz,
            [[3, 1], [2, 2]] * u.Jy, u.MHz, 0.5)   # 3 + x and 2 + 2x
        sefd = model(self.freq).to_value(u.Jy)
        # Pols should give (5, 6) and (6, 8) Jy
        np.testing.assert_allclose(sefd, [np.nan, np.sqrt(30.5), np.sqrt(50.0), np.nan, np.nan])



"""Tests for :mod:`katsdpimager.fast_math`."""

from nose.tools import assert_equal
import numpy as np

from .. import fast_math


class TestExpj2pi:
    def setup(self):
        gen = np.random.Generator(np.random.PCG64(42))
        self.a = gen.random(100) * 1000.0 - 500.0
        self.expected = np.exp(2j * np.pi * self.a)

    def test_basic(self):
        actual = fast_math.expj2pi(self.a)
        np.testing.assert_allclose(actual, self.expected, rtol=1e-12)

    def test_precision(self):
        # Check that precision is respected
        actual = fast_math.expj2pi(self.a.astype(np.float32))
        assert_equal(actual.dtype, np.complex64)
        np.testing.assert_allclose(actual, self.expected, atol=1e-4)


class TestNansum:
    def _test(self, a, *args, **kwargs):
        expected = np.nansum(a, *args, **kwargs)
        actual = fast_math.nansum(a, *args, **kwargs)
        np.testing.assert_array_equal(actual, expected, actual)

    def test_no_nans(self):
        a = np.arange(10.0)
        self._test(a)

    def test_all_nans(self):
        a = np.full(10, np.nan)
        self._test(a)

    def test_some_nans(self):
        a = np.arange(10.0)
        a[::1] = np.nan
        self._test(a)

    def test_2d(self):
        a = np.arange(9.0)
        a[::1] = np.nan
        self._test(a.reshape((3, 3)))
        self._test(a.reshape((3, 3)), axis=0)
        self._test(a.reshape((3, 3)), axis=1)



"""Tests for :py:mod:`katsdpimager.primary_beam`."""

import hashlib

import pkg_resources
import numpy as np
import astropy.units as units
from nose.tools import assert_raises, assert_equal, assert_true

from .. import primary_beam


class TestTrivialBeamModel:
    def setup(self) -> None:
        filename = pkg_resources.resource_filename(
            'katsdpimager',
            'models/beams/meerkat/v1/beam_L.h5')
        self.model = primary_beam.TrivialBeamModel(filename)

    def test_out_of_range(self) -> None:
        freqs = units.Quantity([1.4], units.GHz)
        # Azimuth out of range
        beam = self.model.sample(-0.5, 0.01, 100, 0.0, 0.01, 1, freqs)
        assert_true(np.isnan(beam[0, 0, 0, 0, 0]))
        # Elevation out of range
        beam = self.model.sample(0.0, 0.01, 1, -0.5, 0.01, 100, freqs)
        assert_true(np.isnan(beam[0, 0, 0, 0, 0]))
        # Frequency too low
        beam = self.model.sample(0.0, 0.01, 1, 0.0, 0.01, 1, units.Quantity([0.5], units.GHz))
        assert_true(np.all(np.isnan(beam[0, 0])))
        assert_true(np.all(np.isnan(beam[1, 1])))
        # Frequency too high
        beam = self.model.sample(0.0, 0.01, 1, 0.0, 0.01, 1, units.Quantity([3], units.GHz))
        assert_true(np.all(np.isnan(beam[0, 0])))
        assert_true(np.all(np.isnan(beam[1, 1])))

    def test_parameter_values(self) -> None:
        assert_equal(dict(self.model.parameter_values), {})

    def _check_scalar(self, beam: np.ndarray) -> None:
        """Validate that the beam has no leakage and same H and V response."""
        np.testing.assert_array_equal(beam[1, 0], beam[0, 1])
        np.testing.assert_array_equal(beam[0, 0], beam[1, 1])
        np.testing.assert_array_equal(beam[0, 1], np.zeros_like(beam[0, 1]))

    def _show_beam(self, beam: np.ndarray) -> None:
        """Debug function to plot a beam (should not be any calls checked in)."""
        import matplotlib.pyplot as plt
        rows = 1 if beam.shape[2] == 1 else 2
        fig, axs = plt.subplots(rows, beam.shape[2], sharex=True, sharey=True, squeeze=False)
        for channel in range(beam.shape[2]):
            axs[0][channel].imshow(beam[0, 0, channel], vmin=0, vmax=1.01)
            if channel > 0:
                delta = beam[0, 0, channel] - beam[0, 0, 0]
                axs[1, channel].imshow(delta, vmin=-0.01, vmax=0.01)
        plt.show()

    def test_given_frequency(self) -> None:
        """Sample at frequency already present in the model."""
        N = 101
        STEP = 0.001
        beam = self.model.sample(-STEP * (N // 2), STEP, N,
                                 -STEP * (N // 2), STEP, N,
                                 [1284] * units.MHz)
        assert_equal(beam.shape, (2, 2, 1, N, N))
        self._check_scalar(beam)
        # Should be unity at the centre of the beam
        np.testing.assert_allclose(beam[0, 0, 0, N // 2, N // 2], 1.0, rtol=1e-2)
        # If it fails due to code/model changes, uncomment to inspect visually:
        # self._show_beam(beam)
        assert_equal(hashlib.md5(beam[0, 0, 0]).hexdigest(), '30726bc4a42ddb1813dcdf3a1d54251c')

    def test_interpolated_frequency(self) -> None:
        """Sample at frequency not present in the model."""
        N = 101
        STEP = 0.001
        # Outer two are frequencies in the model
        freqs = [856, 856.41796875, 856.8359375] * units.MHz
        beam = self.model.sample(-STEP * (N // 2), STEP, N,
                                 -STEP * (N // 2), STEP, N,
                                 freqs)
        assert_equal(beam.shape, (2, 2, 3, N, N))
        self._check_scalar(beam)
        # If it fails due to code/model changes, uncomment to inspect visually:
        # self._show_beam(beam)
        assert_equal(hashlib.md5(beam[0, 0, 1]).hexdigest(), '0049684903ec7ce8cb74135fdb17ef15')

    def test_offset(self) -> None:
        N = 128
        STEP = 1 / 2048
        freqs = [1284] * units.MHz
        full = self.model.sample(-STEP * (N // 2), STEP, N,
                                 -STEP * (N // 2), STEP, N,
                                 freqs)
        part = self.model.sample(-STEP * (N // 2), STEP, N // 2,
                                 0, STEP, N // 4,
                                 freqs)
        np.testing.assert_array_equal(part, full[..., 64:96, 0:64])


class TestMeerkatBeamModelSet1:
    def test_load(self):
        """Smoke test to check that the files can be found."""
        for band in primary_beam.MeerkatBeamModelSet1.BANDS:
            models = primary_beam.MeerkatBeamModelSet1(band)
            models.sample()
            assert_equal(models.parameters, [])

    def test_bad_band(self):
        with assert_raises(ValueError):
            primary_beam.MeerkatBeamModelSet1('Z')
